#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
Defines the ``Completer`` class, mother of all shell completion handlers

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2
:status: Alpha

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
from __future__ import absolute_import
import logging

# Try to import readline
try:
    import readline
except ImportError:
    readline = None

# Add some typing
try:
    # pylint: disable=W0611
    from typing import List
    from pelix.framework import BundleContext
    from pelix.shell.beans import ShellSession
    from .decorators import CompletionInfo
except ImportError:
    pass

# Pelix
from pelix.utilities import use_service
from .decorators import SVC_COMPLETER, PROP_COMPLETER_ID, DUMMY

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


class Completer:
    """
    Mother class of all completers
    """

    @staticmethod
    def set_display_hook(display_hook, prompt, session, context):
        try:
            readline.set_completion_display_matches_hook(
                lambda sub, matches, longest: display_hook(
                    prompt, session, context, matches, longest
                )
            )
        except AttributeError:
            # Display hook not available
            pass

    def complete(
        self, config, prompt, session, context, current_arguments, current
    ):
        # type: (CompletionInfo, str, ShellSession, BundleContext, List[str], str) -> List[str]
        """
        Returns the list of bundle IDs matching the current state

        :param config: Configuration of the current completion
        :param prompt: Shell prompt (for re-display)
        :param session: Shell session (to display in shell)
        :param context: Bundle context of the Shell bundle
        :param current_arguments: Current arguments (without the command itself)
        :param current: Current word
        :return: A list of matches
        """
        raise NotImplementedError


# ------------------------------------------------------------------------------


def completion_hints(config, prompt, session, context, current, arguments):
    # type: (CompletionInfo, str, ShellSession, BundleContext, str, List[str]) -> List[str]
    """
    Returns the possible completions of the current argument

    :param config: Configuration of the current completion
    :param prompt: The shell prompt string
    :param session: Current shell session
    :param context: Context of the shell UI bundle
    :param current: Current argument (to be completed)
    :param arguments: List of all arguments in their current state
    :return: A list of possible completions
    """
    if not current:
        # No word yet, so the current position is after the existing ones
        arg_idx = len(arguments)
    else:
        # Find the current word position
        arg_idx = arguments.index(current)

    # Find the ID of the next completer
    completers = config.completers
    if arg_idx > len(completers) - 1:
        # Argument is too far to be positional, try
        if config.multiple:
            # Multiple calls allowed for the last completer
            completer_id = completers[-1]
        else:
            # Nothing to return
            return []
    else:
        completer_id = completers[arg_idx]

    if completer_id == DUMMY:
        # Dummy completer: do nothing
        return []

    # Find the matching service
    svc_ref = context.get_service_reference(
        SVC_COMPLETER, "({}={})".format(PROP_COMPLETER_ID, completer_id)
    )
    if svc_ref is None:
        # Handler not found
        _logger.debug("Unknown shell completer ID: %s", completer_id)
        return []

    # Call the completer
    try:
        with use_service(context, svc_ref) as completer:
            matches = completer.complete(
                config, prompt, session, context, arguments, current
            )
            if not matches:
                return []

            return matches
    except Exception as ex:
        _logger.exception("Error calling completer %s: %s", completer_id, ex)
        return []



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Loads and handles the Pelix initialization file

A configuration file is used to setup a Pelix framework. This module should
be used by shells to load a default configuration.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import json
import os
import sys
import yaml

# Pelix
from pelix.ipopo.constants import use_ipopo
from pelix.utilities import remove_duplicates

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"


# -----------------------------------------------------------------------------


class _Configuration(object):
    """
    Represents a configuration loaded from an initialization file
    """

    def __init__(self):
        """
        Sets up members
        """
        self._properties = {}
        self._environment = {}
        self._paths = []

        self._bundles = []
        self._components = {}

    @property
    def properties(self):
        """
        Returns the configured framework properties

        :return: A dictionary of framework properties
        """
        return self._properties

    @property
    def paths(self):
        """
        Returns the paths to add to sys.path

        :return: A list of paths
        """
        return self._paths

    @property
    def bundles(self):
        """
        Returns the list of bundles to install and start

        :return: A list of names of bundles
        """
        return self._bundles

    @property
    def components(self):
        """
        Returns the definitions of components as a dictionary of tuples.

        Keys are components names, whereas values are (factory, properties)
        tuples

        :return: A name→(factory, properties) dictionary
        """
        return self._components

    def add_properties(self, properties):
        """
        Updates the framework properties dictionary

        :param properties: New framework properties to add
        """
        if isinstance(properties, dict):
            self._properties.update(properties)

    def set_properties(self, properties):
        """
        Sets the framework properties dictionary

        :param properties: Framework properties to set
        """
        self._properties = {}
        self.add_properties(properties)

    def add_environment(self, environ):
        """
        Updates the environment dictionary with the given one.

        Existing entries are overridden by the given ones

        :param environ: New environment variables
        """
        if isinstance(environ, dict):
            self._environment.update(environ)

    def set_environment(self, environ):
        """
        Updates the environment dictionary with the given one.
        Cancels the variables previously set.

        :param environ: New environment variables
        """
        self._environment = {}
        self.add_environment(environ)

    def add_paths(self, paths):
        """
        Adds entries to the Python path.

        The given paths are normalized before being added to the left of the
        list

        :param paths: New paths to add
        """
        if paths:
            # Use new paths in priority
            self._paths = list(paths) + self._paths

    def set_paths(self, paths):
        """
        Adds entries to the Python path.

        The given paths are normalized before being added to the left of the
        Python path.
        Previous paths from configuration files are cleared.

        :param paths: New paths to add
        """
        del self._paths[:]
        self.add_paths(paths)

    def add_bundles(self, bundles):
        """
        Adds a list of bundles to install.

        Contrary to paths and environment variables, the bundles are kept in
        the system-wide to user-specific order.

        :param bundles: A list of bundles to install
        """
        if bundles:
            self._bundles.extend(bundles)

    def set_bundles(self, bundles):
        """
        Adds a list of bundles to install.
        Previous names from configuration files are cleared.

        Contrary to paths and environment variables, the bundles are kept in
        the system-wide to user-specific order.

        :param bundles: A list of bundles to install
        """
        del self._bundles[:]
        self.add_bundles(bundles)

    def add_components(self, components):
        """
        Adds a list of components to instantiate

        :param components: The description of components
        :raise KeyError: Missing component configuration
        """
        if components:
            for component in components:
                self._components[component["name"]] = (
                    component["factory"],
                    component.get("properties", {}),
                )

    def set_components(self, components):
        """
        Adds a list of components to instantiate.
        Removes the previously configured components descriptions.

        :param components: The description of components
        :raise KeyError: Missing component configuration
        """
        self._components.clear()
        self.add_components(components)

    def normalize(self):
        """
        Normalizes environment variables, paths and filters the lists of
        bundles to install and start.

        After this call, the environment variables of this process will have
        been updated.
        """
        # Add environment variables
        os.environ.update(self._environment)

        # Normalize paths and avoid duplicates
        self._paths = remove_duplicates(
            os.path.realpath(os.path.expanduser(os.path.expandvars(path)))
            for path in self._paths
            if os.path.exists(path)
        )

        # Normalize the lists of bundles
        self._bundles = remove_duplicates(self._bundles)


class InitFileHandler(object):
    """
    Parses and handles the instructions of initial configuration files
    """

    DEFAULT_PATH = (
        "/etc/default",
        "/etc",
        "/usr/local/etc",
        "~/.local/pelix",
        "~/.config",
        "~",
        ".",
    )
    """
    Default path where to find the configuration file.
    Order is from system wide to user specific configuration.
    """

    def __init__(self):
        # The internal state
        self.__state = _Configuration()

    @property
    def bundles(self):
        """
        :return: The list of names of bundles to install and start
        """
        return self.__state.bundles

    @property
    def properties(self):
        """
        :return: The initial framework properties
        """
        return self.__state.properties

    def clear(self):
        """
        Clears the current internal state (cleans up all loaded content)
        """
        # Reset the internal state
        self.__state = _Configuration()

    def find_default(self, filename):
        """
        A generate which looks in common folders for the default configuration
        file. The paths goes from system defaults to user specific files.

        :param filename: The name of the file to find
        :return: The complete path to the found files
        """
        for path in self.DEFAULT_PATH:
            # Normalize path
            path = os.path.expanduser(os.path.expandvars(path))
            fullname = os.path.realpath(os.path.join(path, filename))

            if os.path.exists(fullname) and os.path.isfile(fullname):
                yield fullname

    def load(self, filename=None):
        """
        Loads the given file and adds its content to the current state.
        This method can be called multiple times to merge different files.

        If no filename is given, this method loads all default files found.
        It returns False if no default configuration file has been found

        :param filename: The file to load
        :return: True if the file has been correctly parsed, False if no file
                 was given and no default file exist
        :raise IOError: Error loading file
        """
        if not filename:
            for name in self.find_default(".pelix.conf"):
                self.load(name)
        else:
            _, file_extension = os.path.splitext(filename)
            with open(filename, "r") as filep:
                if file_extension == ".yaml" or file_extension == ".yml":
                    self.__parse(yaml.safe_load(filep))
                else:
                    self.__parse(json.load(filep))

    def __parse(self, configuration):
        """
        Parses the given configuration dictionary

        :param configuration: A configuration as a dictionary (JSON object)
        """
        for entry in (
            "properties",
            "environment",
            "paths",
            "bundles",
            "components",
        ):
            # Check if current values must be reset
            reset_key = "reset_{0}".format(entry)

            # Compute the name of the method
            call_name = "add" if not configuration.get(reset_key) else "set"
            method = getattr(self.__state, "{0}_{1}".format(call_name, entry))

            # Update configuration
            method(configuration.get(entry))

    def normalize(self):
        """
        Normalizes environment variables and the Python path.

        This method first updates the environment variables (``os.environ``).
        Then, it normalizes the Python path (``sys.path``) by resolving all
        references to the user directory and environment variables.
        """
        # Normalize configuration
        self.__state.normalize()

        # Update sys.path, avoiding duplicates
        whole_path = list(self.__state.paths)
        whole_path.extend(sys.path)

        # Ensure the working directory as first search path
        sys.path = ["."]
        for path in whole_path:
            if path not in sys.path:
                sys.path.append(path)

    def instantiate_components(self, context):
        """
        Instantiate the defined components

        .. note::
           This method requires the iPOPO core service to be registered.
           This means that the ``pelix.ipopo.core`` must have been declared in
           the list of bundles (or installed and started programmatically).

        :param context: A :class:`~pelix.framework.BundleContext` object
        :raise BundleException: Error looking for the iPOPO service or
                                starting a component
        """
        with use_ipopo(context) as ipopo:
            for name, (factory, properties) in self.__state.components.items():
                ipopo.instantiate(factory, name, properties)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
RequiresBest handler implementation

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Pelix beans
from pelix.constants import BundleActivator, SERVICE_RANKING

# iPOPO constants
import pelix.ipopo.constants as ipopo_constants
import pelix.ipopo.handlers.constants as constants
import pelix.ipopo.handlers.requires as requires

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _HandlerFactory(requires._HandlerFactory):
    # pylint: disable=W0212, R0903
    """
    Factory service for service registration handlers
    """

    def get_handlers(self, component_context, instance):
        """
        Sets up service providers for the given component

        :param component_context: The ComponentContext bean
        :param instance: The component instance
        :return: The list of handlers associated to the given component
        """
        # Extract information from the context
        requirements = component_context.get_handler(
            ipopo_constants.HANDLER_REQUIRES_BEST
        )
        requires_filters = component_context.properties.get(
            ipopo_constants.IPOPO_REQUIRES_FILTERS, None
        )

        # Prepare requirements
        requirements = self._prepare_requirements(
            requirements, requires_filters
        )

        # Set up the runtime dependency handlers
        return [
            BestDependency(field, requirement)
            for field, requirement in requirements.items()
        ]


@BundleActivator
class _Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Sets up members
        """
        self._registration = None

    def start(self, context):
        """
        Bundle started
        """
        # Set up properties
        properties = {
            constants.PROP_HANDLER_ID: ipopo_constants.HANDLER_REQUIRES_BEST
        }

        # Register the handler factory service
        self._registration = context.register_service(
            constants.SERVICE_IPOPO_HANDLER_FACTORY,
            _HandlerFactory(),
            properties,
        )

    def stop(self, _):
        """
        Bundle stopped
        """
        # Unregister the service
        self._registration.unregister()
        self._registration = None


# ------------------------------------------------------------------------------


class BestDependency(requires.SimpleDependency):
    """
    Manages a simple dependency field

    TODO: Allow to use a custom service reference comparator
    """

    def __init__(self, field, requirement):
        """
        Sets up members
        """
        super(BestDependency, self).__init__(field, requirement)

        # Current ranking
        self._current_ranking = None

    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        """
        self._current_ranking = None
        super(BestDependency, self).clear()

    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            new_ranking = svc_ref.get_property(SERVICE_RANKING)
            if self._current_ranking is not None:
                if new_ranking > self._current_ranking:
                    # New service with better ranking: use it
                    self._pending_ref = svc_ref
                    old_ref = self.reference
                    old_value = self._value

                    # Clean up like for a departure
                    self._current_ranking = None
                    self._value = None
                    self.reference = None

                    # Unbind (new binding will be done afterwards)
                    self._ipopo_instance.unbind(self, old_value, old_ref)
            else:
                # No ranking yet: inject the service
                self.reference = svc_ref
                self._value = self._context.get_service(svc_ref)
                self._current_ranking = new_ranking
                self._pending_ref = None

                self._ipopo_instance.bind(self, self._value, self.reference)

    def on_service_departure(self, svc_ref):
        """
        Called when a service has been unregistered from the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            if svc_ref is self.reference:
                # Injected service going away...
                service = self._value

                # Clear the instance values
                self._current_ranking = None
                self._value = None
                self.reference = None

                if self.requirement.immediate_rebind:
                    # Look for a replacement
                    self._pending_ref = self._context.get_service_reference(
                        self.requirement.specification, self.requirement.filter
                    )
                else:
                    self._pending_ref = None

                self._ipopo_instance.unbind(self, service, svc_ref)

    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been modified in the framework

        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        """
        with self._lock:
            if self.reference is None:
                # A previously registered service now matches our filter
                return self.on_service_arrival(svc_ref)
            else:
                # Check if the ranking changed the service to inject
                best_ref = self._context.get_service_reference(
                    self.requirement.specification, self.requirement.filter
                )
                if best_ref is self.reference:
                    # Still the best service: notify the property modification
                    if svc_ref is self.reference:
                        # Call update only if necessary
                        self._ipopo_instance.update(
                            self, self._value, svc_ref, old_properties
                        )
                else:
                    # A new service is now the best: start a departure loop
                    self.on_service_departure(self.reference)

            return None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Core module for Pelix.

Pelix is a Python framework that aims to act as OSGi as much as possible

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import collections
import importlib
import inspect
import logging
import os
import sys
import threading
import uuid

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, List, Optional, Set, Union
    import types
except ImportError:
    pass

# Pelix beans and constants
from pelix.constants import (
    ACTIVATOR,
    ACTIVATOR_LEGACY,
    FRAMEWORK_UID,
    OSGI_FRAMEWORK_UUID,
    BundleException,
    FrameworkException,
)
from pelix.internals.events import BundleEvent, ServiceEvent

# pylint: disable=W0611
from pelix.internals.registry import (
    EventDispatcher,
    ServiceRegistry,
    ServiceReference,
    ServiceRegistration,
)

# Pelix utility modules
from pelix.utilities import is_string


if hasattr(importlib, "reload"):
    # This method has been added in Python 3.4 and deprecates imp.reload()
    def reload_module(module_):
        """
        Reloads a module using ``importlib.reload()`` when available

        :param module_: The module to update
        :return: The new version of the module
        :raise ImportError: Error looking for file
        :raise SyntaxError: Syntax error in imported module
        """
        return importlib.reload(module_)


else:
    # Before Python 3.4
    import imp

    def reload_module(module_):
        """
        Reloads a module using ``imp.reload()`` as fallback

        :param module_: The module to update
        :return: The new version of the module
        :raise ImportError: Error looking for file
        :raise SyntaxError: Syntax error in imported module
        """
        return imp.reload(module_)


def walk_modules(path):
    """
    Code from ``pkgutil.ImpImporter.iter_modules()``: walks through a folder
    and yields all loadable packages and modules.

    :param path: Path where to look for modules
    :return: Generator to walk through found packages and modules
    """
    if path is None or not os.path.isdir(path):
        return

    yielded = set()
    try:
        file_names = os.listdir(path)
    except OSError:
        # ignore unreadable directories like import does
        file_names = []

    # handle packages before same-named modules
    file_names.sort()

    for filename in file_names:
        modname = inspect.getmodulename(filename)
        if modname == "__init__" or modname in yielded:
            continue

        file_path = os.path.join(path, filename)
        is_package = False

        if not modname and os.path.isdir(file_path) and "." not in filename:
            modname = filename
            try:
                dir_contents = os.listdir(file_path)
            except OSError:
                # ignore unreadable directories like import does
                dir_contents = []

            for sub_filename in dir_contents:
                sub_name = inspect.getmodulename(sub_filename)
                if sub_name == "__init__":
                    is_package = True
                    break
            else:
                # not a package
                continue

        if modname and "." not in modname:
            yielded.add(modname)
            yield modname, is_package


# ------------------------------------------------------------------------------


# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Prepare the module logger
_logger = logging.getLogger("pelix.main")

# ------------------------------------------------------------------------------


class Bundle(object):
    # pylint: disable=W0212
    """
    Represents a "bundle" in Pelix
    """

    __slots__ = (
        "_lock",
        "__context",
        "__id",
        "__module",
        "__name",
        "__framework",
        "_state",
        "__registered_services",
        "__registration_lock",
    )

    UNINSTALLED = 1
    """ The bundle is uninstalled and may not be used """

    INSTALLED = 2
    """ The bundle is installed but not yet resolved """

    RESOLVED = 4
    """ The bundle is resolved and is able to be started """

    STARTING = 8
    """ The bundle is in the process of starting """

    STOPPING = 16
    """ The bundle is in the process of stopping """

    ACTIVE = 32
    """ The bundle is now running """

    def __init__(self, framework, bundle_id, name, module_):
        # type: (Framework, int, str, types.ModuleType) -> None
        """
        Sets up the bundle descriptor

        :param framework: The host framework
        :param bundle_id: The ID of the bundle in the host framework
        :param name: The bundle symbolic name
        :param module_: The bundle module
        """
        # A re-entrant lock for synchronization
        self._lock = threading.RLock()

        # Bundle
        self.__context = BundleContext(framework, self)
        self.__id = bundle_id
        self.__module = module_
        self.__name = name

        self.__framework = framework
        self._state = Bundle.RESOLVED

        # Registered services
        self.__registered_services = set()  # type: Set[ServiceRegistration]
        self.__registration_lock = threading.Lock()

    def __str__(self):
        """
        String representation
        """
        return "Bundle(ID={0}, Name={1})".format(self.__id, self.__name)

    def __get_activator_method(self, method_name):
        """
        Retrieves the requested method of the activator, or returns None

        :param method_name: A method name
        :return: A method, or None
        """
        # Get the activator
        activator = getattr(self.__module, ACTIVATOR, None)
        if activator is None:
            # Get the old activator
            activator = getattr(self.__module, ACTIVATOR_LEGACY, None)
            if activator is not None:
                # Old activator found: print a deprecation warning
                _logger.warning(
                    "Bundle %s uses the deprecated '%s' to declare"
                    " its activator. Use @BundleActivator instead.",
                    self.__name,
                    ACTIVATOR_LEGACY,
                )
        return getattr(activator, method_name, None)

    def _fire_bundle_event(self, kind):
        # type: (int) -> None
        """
        Fires a bundle event of the given kind

        :param kind: Kind of event
        """
        self.__framework._dispatcher.fire_bundle_event(BundleEvent(kind, self))

    def _registered_service(self, registration):
        # type: (ServiceRegistration) -> None
        """
        Bundle is notified by the framework that a service has been registered
        in the name of this bundle.

        :param registration: The service registration object
        """
        with self.__registration_lock:
            self.__registered_services.add(registration)

    def _unregistered_service(self, registration):
        # type: (ServiceRegistration) -> None
        """
        Bundle is notified by the framework that a service has been
        unregistered in the name of this bundle.

        :param registration: The service registration object
        """
        with self.__registration_lock:
            self.__registered_services.discard(registration)

    def get_bundle_context(self):
        # type: () -> BundleContext
        """
        Retrieves the bundle context

        :return: The bundle context
        """
        return self.__context

    def get_bundle_id(self):
        # type: () -> int
        """
        Retrieves the bundle ID

        :return: The bundle ID
        """
        return self.__id

    def get_location(self):
        # type: () -> str
        """
        Retrieves the location of this module

        :return: The location of the Pelix module, or an empty string
        """
        return getattr(self.__module, "__file__", "")

    def get_module(self):

        # type: () -> types.ModuleType
        """
        Retrieves the Python module corresponding to the bundle

        :return: The Python module
        """
        return self.__module

    def get_registered_services(self):
        # type: () -> List[ServiceReference]
        """
        Returns this bundle's ServiceReference list for all services it has
        registered or an empty list

        The list is valid at the time of the call to this method, however, as
        the Framework is a very dynamic environment, services can be modified
        or unregistered at any time.

        :return: An array of ServiceReference objects
        :raise BundleException: If the bundle has been uninstalled
        """
        if self._state == Bundle.UNINSTALLED:
            raise BundleException(
                "Can't call 'get_registered_services' on an "
                "uninstalled bundle"
            )
        return self.__framework._registry.get_bundle_registered_services(self)

    def get_services_in_use(self):
        # type: () -> List[ServiceReference]
        """
        Returns this bundle's ServiceReference list for all services it is
        using or an empty list.
        A bundle is considered to be using a service if its use count for that
        service is greater than zero.

        The list is valid at the time of the call to this method, however, as
        the Framework is a very dynamic environment, services can be modified
        or unregistered at any time.

        :return: An array of ServiceReference objects
        :raise BundleException: If the bundle has been uninstalled
        """
        if self._state == Bundle.UNINSTALLED:
            raise BundleException(
                "Can't call 'get_services_in_use' on an uninstalled bundle"
            )
        return self.__framework._registry.get_bundle_imported_services(self)

    def get_state(self):
        # type: () -> int
        """
        Retrieves the bundle state

        :return: The bundle state
        """
        return self._state

    def get_symbolic_name(self):
        # type: () -> str
        """
        Retrieves the bundle symbolic name (its Python module name)

        :return: The bundle symbolic name
        """
        return self.__name

    def get_version(self):
        # type: () -> str
        """
        Retrieves the bundle version, using the ``__version__`` or
        ``__version_info__`` attributes of its module.

        :return: The bundle version, "0.0.0" by default
        """
        # Get the version value
        version = getattr(self.__module, "__version__", None)
        if version:
            return version

        # Convert the __version_info__ entry
        info = getattr(self.__module, "__version_info__", None)
        if info:
            return ".".join(str(part) for part in __version_info__)

        # No version
        return "0.0.0"

    def start(self):
        """
        Starts the bundle. Does nothing if the bundle is already starting or
        active.

        :raise BundleException: The framework is not yet started or the bundle
                                activator failed.
        """
        if self.__framework._state not in (Bundle.STARTING, Bundle.ACTIVE):
            # Framework is not running
            raise BundleException(
                "Framework must be started before its bundles"
            )

        with self._lock:
            if self._state in (Bundle.ACTIVE, Bundle.STARTING):
                # Already started bundle, do nothing
                return

            # Store the bundle current state
            previous_state = self._state

            # Starting...
            self._state = Bundle.STARTING
            self._fire_bundle_event(BundleEvent.STARTING)

            # Call the activator, if any
            starter = self.__get_activator_method("start")
            if starter is not None:
                try:
                    # Call the start method
                    starter(self.__context)
                except (FrameworkException, BundleException):
                    # Restore previous state
                    self._state = previous_state

                    # Re-raise directly Pelix exceptions
                    _logger.exception(
                        "Pelix error raised by %s while starting", self.__name
                    )
                    raise
                except Exception as ex:
                    # Restore previous state
                    self._state = previous_state

                    # Raise the error
                    _logger.exception(
                        "Error raised by %s while starting", self.__name
                    )
                    raise BundleException(ex)

            # Bundle is now active
            self._state = Bundle.ACTIVE
            self._fire_bundle_event(BundleEvent.STARTED)

    def stop(self):
        """
        Stops the bundle. Does nothing if the bundle is already stopped.

        :raise BundleException: The bundle activator failed.
        """
        if self._state != Bundle.ACTIVE:
            # Invalid state
            return

        exception = None
        with self._lock:
            # Store the bundle current state
            previous_state = self._state

            # Stopping...
            self._state = Bundle.STOPPING
            self._fire_bundle_event(BundleEvent.STOPPING)

            # Call the activator, if any
            stopper = self.__get_activator_method("stop")
            if stopper is not None:
                try:
                    # Call the start method
                    stopper(self.__context)
                except (FrameworkException, BundleException) as ex:
                    # Restore previous state
                    self._state = previous_state

                    # Re-raise directly Pelix exceptions
                    _logger.exception(
                        "Pelix error raised by %s while stopping", self.__name
                    )
                    exception = ex
                except Exception as ex:
                    _logger.exception(
                        "Error raised by %s while stopping", self.__name
                    )
                    # Store the exception (raised after service clean up)
                    exception = BundleException(ex)

            # Hide remaining services
            self.__framework._hide_bundle_services(self)

            # Intermediate bundle event : activator should have cleaned up
            # everything, but some element could stay (iPOPO components, ...)
            self._fire_bundle_event(BundleEvent.STOPPING_PRECLEAN)

            # Remove remaining services (the hard way)
            self.__unregister_services()

            # Cleanup service usages
            self.__framework._unget_used_services(self)

            # Bundle is now stopped and all its services have been unregistered
            self._state = Bundle.RESOLVED
            self._fire_bundle_event(BundleEvent.STOPPED)

        # Raise the exception, if any
        # pylint: disable=E0702
        # Pylint seems to miss the "is not None" check below
        if exception is not None:
            raise exception

    def __unregister_services(self):
        """
        Unregisters all bundle services
        """
        # Copy the services list, as it will be modified during the process
        with self.__registration_lock:
            registered_services = self.__registered_services.copy()

        for registration in registered_services:
            try:
                registration.unregister()
            except BundleException:
                # Ignore errors at this level
                pass

        if self.__registered_services:
            _logger.warning("Not all services have been unregistered...")

        with self.__registration_lock:
            # Clear the list, just to be clean
            self.__registered_services.clear()

    def uninstall(self):
        """
        Uninstalls the bundle
        """
        with self._lock:
            if self._state == Bundle.ACTIVE:
                self.stop()

            # Change the bundle state
            self._state = Bundle.UNINSTALLED

            # Call the framework
            self.__framework.uninstall_bundle(self)

    def update(self):
        """
        Updates the bundle
        """
        with self._lock:
            # Was it active ?
            restart = self._state == Bundle.ACTIVE

            # Send the update event
            self._fire_bundle_event(BundleEvent.UPDATE_BEGIN)

            try:
                # Stop the bundle
                self.stop()
            except:
                # Something wrong occurred, notify listeners
                self._fire_bundle_event(BundleEvent.UPDATE_FAILED)
                raise

            # Change the source file age
            module_stat = None
            module_file = getattr(self.__module, "__file__", None)
            if module_file is not None and os.path.isfile(module_file):
                try:
                    module_stat = os.stat(module_file)

                    # Change modification time to bypass weak time resolution
                    # of the underlying file system
                    os.utime(
                        module_file,
                        (module_stat.st_atime, module_stat.st_mtime + 1),
                    )
                except OSError:
                    # Can't touch the file
                    _logger.warning(
                        "Failed to update the modification time of '%s'. "
                        "The bundle update might not reflect the latest "
                        "changes.",
                        module_file,
                    )

            # Clean up the module constants (otherwise kept by reload)
            # Keep special members (__name__, __file__, ...)
            old_content = self.__module.__dict__.copy()
            for name in list(self.__module.__dict__):
                if not (name.startswith("__") and name.endswith("__")):
                    del self.__module.__dict__[name]

            try:
                # Reload the module
                reload_module(self.__module)
            except (ImportError, SyntaxError) as ex:
                # Exception raised if the file is unreadable
                _logger.exception("Error updating %s: %s", self.__name, ex)

                # Reset module content
                self.__module.__dict__.clear()
                self.__module.__dict__.update(old_content)

            if module_stat is not None:
                try:
                    # Reset times
                    os.utime(
                        module_file,
                        (module_stat.st_atime, module_stat.st_mtime),
                    )
                except OSError:
                    # Shouldn't occur, since we succeeded before the update
                    _logger.debug(
                        "Failed to reset the modification time of '%s'",
                        module_file,
                    )

            if restart:
                try:
                    # Re-start the bundle
                    self.start()
                except:
                    # Something wrong occurred, notify listeners
                    self._fire_bundle_event(BundleEvent.UPDATE_FAILED)
                    raise

            # Bundle update finished
            self._fire_bundle_event(BundleEvent.UPDATED)


# ------------------------------------------------------------------------------


class Framework(Bundle):
    # pylint: disable=W0212
    """
    The Pelix framework (main) class. It must be instantiated using
    FrameworkFactory
    """

    def __init__(self, properties=None):
        """
        Sets up the framework.

        :param properties: The framework properties
        """
        # Framework bundle set up
        Bundle.__init__(
            self, self, 0, self.get_symbolic_name(), sys.modules[__name__]
        )

        # Framework properties
        if not isinstance(properties, dict):
            self.__properties = {}
        else:
            # Use a copy of the properties, to avoid external changes
            self.__properties = properties.copy()

        # Generate and set a framework instance UUID, if needed
        framework_uid = self.__properties.get(FRAMEWORK_UID)
        if not framework_uid:
            framework_uid = str(uuid.uuid4())
        # Normalize the UID: it must be a string
        self.__properties[FRAMEWORK_UID] = str(framework_uid)
        # Also normalize the OSGI_FRAMEWORK_UID: it must be a string
        self.__properties[OSGI_FRAMEWORK_UUID] = str(framework_uid)

        # Properties lock
        self.__properties_lock = threading.Lock()

        # Bundles (start at 1, as 0 is reserved for the framework itself)
        self.__next_bundle_id = 1

        # Bundle ID -> Bundle object
        self.__bundles = {}

        # Bundles lock
        self.__bundles_lock = threading.RLock()

        # Service registry
        self._registry = ServiceRegistry(self)
        self.__unregistering_services = {}

        # Event dispatcher
        self._dispatcher = EventDispatcher(self._registry)

        # The wait_for_stop event (initially stopped)
        self._fw_stop_event = threading.Event()
        self._fw_stop_event.set()

    def add_property(self, name, value):
        # type: (str, object) -> bool
        """
        Adds a property to the framework **if it is not yet set**.

        If the property already exists (same name), then nothing is done.
        Properties can't be updated.

        :param name: The property name
        :param value: The value to set
        :return: True if the property was stored, else False
        """
        with self.__properties_lock:
            if name in self.__properties:
                # Already stored property
                return False

            self.__properties[name] = value
            return True

    def find_service_references(
        self, clazz=None, ldap_filter=None, only_one=False
    ):
        # type: (Optional[str], Optional[str], bool) -> Optional[List[ServiceReference]]
        """
        Finds all services references matching the given filter.

        :param clazz: Class implemented by the service
        :param ldap_filter: Service filter
        :param only_one: Return the first matching service reference only
        :return: A list of found reference, or None
        :raise BundleException: An error occurred looking for service
                                references
        """
        return self._registry.find_service_references(
            clazz, ldap_filter, only_one
        )

    def get_bundle_by_id(self, bundle_id):
        # type: (int) -> Union[Bundle, Framework]
        """
        Retrieves the bundle with the given ID

        :param bundle_id: ID of an installed bundle
        :return: The requested bundle
        :raise BundleException: The ID is invalid
        """
        if bundle_id == 0:
            # "System bundle"
            return self

        with self.__bundles_lock:
            if bundle_id not in self.__bundles:
                raise BundleException("Invalid bundle ID {0}".format(bundle_id))

            return self.__bundles[bundle_id]

    def get_bundle_by_name(self, bundle_name):
        # type: (str) -> Optional[Bundle]
        """
        Retrieves the bundle with the given name

        :param bundle_name: Name of the bundle to look for
        :return: The requested bundle, None if not found
        """
        if bundle_name is None:
            # Nothing to do
            return None

        if bundle_name is self.get_symbolic_name():
            # System bundle requested
            return self

        with self.__bundles_lock:
            for bundle in self.__bundles.values():
                if bundle_name == bundle.get_symbolic_name():
                    # Found !
                    return bundle

            # Not found...
            return None

    def get_bundles(self):
        # type: () -> List[Bundle]
        """
        Returns the list of all installed bundles

        :return: the list of all installed bundles
        """
        with self.__bundles_lock:
            return [
                self.__bundles[bundle_id]
                for bundle_id in sorted(self.__bundles.keys())
            ]

    def get_properties(self):
        # type: () -> dict
        """
        Retrieves a copy of the stored framework properties.
        """
        with self.__properties_lock:
            return self.__properties.copy()

    def get_property(self, name):
        # type: (str) -> object
        """
        Retrieves a framework or system property. As framework properties don't
        change while it's running, this method don't need to be protected.

        :param name: The property name
        """
        with self.__properties_lock:
            return self.__properties.get(name, os.getenv(name))

    def get_property_keys(self):
        # type: () -> tuple
        """
        Returns an array of the keys in the properties of the service

        :return: An array of property keys.
        """
        with self.__properties_lock:
            return tuple(self.__properties.keys())

    def get_service(self, bundle, reference):
        # type: (Bundle, ServiceReference) -> Any
        """
        Retrieves the service corresponding to the given reference

        :param bundle: The bundle requiring the service
        :param reference: A service reference
        :return: The requested service
        :raise BundleException: The service could not be found
        :raise TypeError: The argument is not a ServiceReference object
        """
        if not isinstance(bundle, Bundle):
            raise TypeError("First argument must be a Bundle object")
        elif not isinstance(reference, ServiceReference):
            raise TypeError("Second argument must be a ServiceReference object")

        try:
            # Unregistering service, just give it
            return self.__unregistering_services[reference]
        except KeyError:
            return self._registry.get_service(bundle, reference)

    def _get_service_objects(self, bundle, reference):
        # type: (Bundle, ServiceReference) -> ServiceObjects
        """
        Returns the ServiceObjects object for the service referenced by the
        specified ServiceReference object.

        :param bundle: The bundle requiring the service
        :param reference: Reference to a prototype service factory
        :return: An intermediate object to get more instances of a service
        """
        return ServiceObjects(self._registry, bundle, reference)

    def get_symbolic_name(self):
        # type: () -> str
        """
        Retrieves the framework symbolic name

        :return: Always "pelix.framework"
        """
        return "pelix.framework"

    def install_bundle(self, name, path=None):
        # type: (str, str) -> Bundle
        """
        Installs the bundle with the given name

        *Note:* Before Pelix 0.5.0, this method returned the ID of the
        installed bundle, instead of the Bundle object.

        **WARNING:** The behavior of the loading process is subject to changes,
        as it does not allow to safely run multiple frameworks in the same
        Python interpreter, as they might share global module values.

        :param name: A bundle name
        :param path: Preferred path to load the module
        :return: The installed Bundle object
        :raise BundleException: Something happened
        """
        with self.__bundles_lock:
            # A bundle can't be installed twice
            for bundle in self.__bundles.values():
                if bundle.get_symbolic_name() == name:
                    _logger.debug("Already installed bundle: %s", name)
                    return bundle

            # Load the module
            try:
                if path:
                    # Use the given path in priority
                    sys.path.insert(0, path)

                try:
                    # The module has already been loaded
                    module_ = sys.modules[name]
                except KeyError:
                    # Load the module
                    #  __import__(name) -> package level
                    # import_module -> module level
                    module_ = importlib.import_module(name)
            except (ImportError, IOError) as ex:
                # Error importing the module
                raise BundleException(
                    "Error installing bundle {0}: {1}".format(name, ex)
                )
            finally:
                if path:
                    # Clean up the path. The loaded module(s) might
                    # have changed the path content, so do not use an
                    # index
                    sys.path.remove(path)

            # Add the module to sys.modules, just to be sure
            sys.modules[name] = module_

            # Compute the bundle ID
            bundle_id = self.__next_bundle_id

            # Prepare the bundle object and its context
            bundle = Bundle(self, bundle_id, name, module_)

            # Store the bundle
            self.__bundles[bundle_id] = bundle

            # Update the bundle ID counter
            self.__next_bundle_id += 1

        # Fire the bundle installed event
        event = BundleEvent(BundleEvent.INSTALLED, bundle)
        self._dispatcher.fire_bundle_event(event)
        return bundle

    def install_package(self, path, recursive=False, prefix=None):
        # type: (str, bool, str) -> tuple
        """
        Installs all the modules found in the given package

        :param path: Path of the package (folder)
        :param recursive: If True, install the sub-packages too
        :param prefix: (**internal**) Prefix for all found modules
        :return: A 2-tuple, with the list of installed bundles and the list
                 of failed modules names
        :raise ValueError: Invalid path
        """
        if not path:
            raise ValueError("Empty path")
        elif not is_string(path):
            raise ValueError("Path must be a string")

        # Use an absolute path
        path = os.path.abspath(path)
        if not os.path.exists(path):
            raise ValueError("Nonexistent path: {0}".format(path))

        # Create a simple visitor
        def visitor(fullname, is_package, module_path):
            # pylint: disable=W0613
            """
            Package visitor: accepts everything in recursive mode,
            else avoids packages
            """
            return recursive or not is_package

        # Set up the prefix if needed
        if prefix is None:
            prefix = os.path.basename(path)

        bundles = set()  # type: Set[Bundle]
        failed = set()  # type: Set[str]

        with self.__bundles_lock:
            try:
                # Install the package first, resolved from the parent directory
                bundles.add(self.install_bundle(prefix, os.path.dirname(path)))

                # Visit the package
                visited, sub_failed = self.install_visiting(
                    path, visitor, prefix
                )

                # Update the sets
                bundles.update(visited)
                failed.update(sub_failed)
            except BundleException as ex:
                # Error loading the module
                _logger.warning("Error loading package %s: %s", prefix, ex)
                failed.add(prefix)

        return bundles, failed

    def install_visiting(self, path, visitor, prefix=None):
        """
        Installs all the modules found in the given path if they are accepted
        by the visitor.

        The visitor must be a callable accepting 3 parameters:

           * fullname: The full name of the module
           * is_package: If True, the module is a package
           * module_path: The path to the module file

        :param path: Root search path
        :param visitor: The visiting callable
        :param prefix: (**internal**) Prefix for all found modules
        :return: A 2-tuple, with the list of installed bundles and the list
                 of failed modules names
        :raise ValueError: Invalid path or visitor
        """
        # Validate the path
        if not path:
            raise ValueError("Empty path")
        elif not is_string(path):
            raise ValueError("Path must be a string")

        # Validate the visitor
        if visitor is None:
            raise ValueError("No visitor method given")

        # Use an absolute path
        path = os.path.abspath(path)
        if not os.path.exists(path):
            raise ValueError("Inexistent path: {0}".format(path))

        # Set up the prefix if needed
        if prefix is None:
            prefix = os.path.basename(path)

        bundles = set()
        failed = set()

        with self.__bundles_lock:
            # Walk through the folder to find modules
            for name, is_package in walk_modules(path):
                # Ignore '__main__' modules
                if name == "__main__":
                    continue

                # Compute the full name of the module
                fullname = ".".join((prefix, name)) if prefix else name
                try:
                    if visitor(fullname, is_package, path):
                        if is_package:
                            # Install the package
                            bundles.add(self.install_bundle(fullname, path))

                            # Visit the package
                            sub_path = os.path.join(path, name)
                            sub_bundles, sub_failed = self.install_visiting(
                                sub_path, visitor, fullname
                            )
                            bundles.update(sub_bundles)
                            failed.update(sub_failed)
                        else:
                            # Install the bundle
                            bundles.add(self.install_bundle(fullname, path))
                except BundleException as ex:
                    # Error loading the module
                    _logger.warning("Error visiting %s: %s", fullname, ex)

                    # Try the next module
                    failed.add(fullname)
                    continue

        return bundles, failed

    def register_service(
        self,
        bundle,
        clazz,
        service,
        properties,
        send_event,
        factory=False,
        prototype=False,
    ):
        # type: (Bundle, Union[List[Any], type, str], object, dict, bool, bool, bool) -> ServiceRegistration
        """
        Registers a service and calls the listeners

        :param bundle: The bundle registering the service
        :param clazz: Name(s) of the interface(s) implemented by service
        :param service: The service to register
        :param properties: Service properties
        :param send_event: If not, doesn't trigger a service registered event
        :param factory: If True, the given service is a service factory
        :param prototype: If True, the given service is a prototype service
                          factory (the factory argument is considered True)
        :return: A ServiceRegistration object
        :raise BundleException: An error occurred while registering the service
        """
        if bundle is None or service is None or not clazz:
            raise BundleException("Invalid registration parameters")

        if not isinstance(properties, dict):
            # Be sure we have a valid dictionary
            properties = {}
        else:
            # Use a copy of the given properties
            properties = properties.copy()

        # Prepare the class specification
        if not isinstance(clazz, (list, tuple)):
            # Make a list from the single class
            clazz = [clazz]

        # Test the list content
        classes = []
        for svc_clazz in clazz:
            if inspect.isclass(svc_clazz):
                # Keep the type name
                svc_clazz = svc_clazz.__name__

            if not svc_clazz or not is_string(svc_clazz):
                # Invalid class name
                raise BundleException(
                    "Invalid class name: {0}".format(svc_clazz)
                )

            # Class OK
            classes.append(svc_clazz)

        # Make the service registration
        registration = self._registry.register(
            bundle, classes, properties, service, factory, prototype
        )

        # Update the bundle registration information
        bundle._registered_service(registration)

        if send_event:
            # Call the listeners
            event = ServiceEvent(
                ServiceEvent.REGISTERED, registration.get_reference()
            )
            self._dispatcher.fire_service_event(event)

        return registration

    def start(self):
        # type: () -> bool
        """
        Starts the framework

        :return: True if the bundle has been started, False if it was already
                 running
        :raise BundleException: A bundle failed to start
        """
        with self._lock:
            if self._state in (Bundle.STARTING, Bundle.ACTIVE):
                # Already started framework
                return False

            # Reset the stop event
            self._fw_stop_event.clear()

            # Starting...
            self._state = Bundle.STARTING
            self._dispatcher.fire_bundle_event(
                BundleEvent(BundleEvent.STARTING, self)
            )

            # Start all registered bundles (use a copy, just in case...)
            for bundle in self.__bundles.copy().values():
                try:
                    bundle.start()
                except FrameworkException as ex:
                    # Important error
                    _logger.exception(
                        "Important error starting bundle: %s", bundle
                    )
                    if ex.needs_stop:
                        # Stop the framework (has to be in active state)
                        self._state = Bundle.ACTIVE
                        self.stop()
                        return False
                except BundleException:
                    # A bundle failed to start : just log
                    _logger.exception("Error starting bundle: %s", bundle)

            # Bundle is now active
            self._state = Bundle.ACTIVE
            return True

    def stop(self):
        # type: () -> bool
        """
        Stops the framework

        :return: True if the framework stopped, False it wasn't running
        """
        with self._lock:
            if self._state != Bundle.ACTIVE:
                # Invalid state
                return False

            # Hide all services (they will be deleted by bundle.stop())
            for bundle in self.__bundles.values():
                self._registry.hide_bundle_services(bundle)

            # Stopping...
            self._state = Bundle.STOPPING
            self._dispatcher.fire_bundle_event(
                BundleEvent(BundleEvent.STOPPING, self)
            )

            # Notify listeners that the bundle is stopping
            self._dispatcher.fire_framework_stopping()

            bid = self.__next_bundle_id - 1
            while bid > 0:
                bundle = self.__bundles.get(bid)
                bid -= 1

                if bundle is None or bundle.get_state() != Bundle.ACTIVE:
                    # Ignore inactive bundle
                    continue

                try:
                    bundle.stop()
                except Exception as ex:
                    # Just log exceptions
                    _logger.exception(
                        "Error stopping bundle %s: %s",
                        bundle.get_symbolic_name(),
                        ex,
                    )

            # Framework is now stopped
            self._state = Bundle.RESOLVED
            self._dispatcher.fire_bundle_event(
                BundleEvent(BundleEvent.STOPPED, self)
            )

            # All bundles have been stopped, release "wait_for_stop"
            self._fw_stop_event.set()

            # Force the registry clean up
            self._registry.clear()
            return True

    def delete(self, force=False):
        """
        Deletes the current framework

        :param force: If True, stops the framework before deleting it
        :return: True if the framework has been delete, False if is couldn't
        """
        if not force and self._state not in (
            Bundle.INSTALLED,
            Bundle.RESOLVED,
            Bundle.STOPPING,
        ):
            _logger.warning("Trying to delete an active framework")
            return False

        return FrameworkFactory.delete_framework(self)

    def uninstall(self):
        """
        A framework can't be uninstalled

        :raise BundleException: This method must not be called
        """
        raise BundleException("A framework can't be uninstalled")

    def uninstall_bundle(self, bundle):
        # type: (Bundle) -> None
        """
        Ends the uninstallation of the given bundle (must be called by Bundle)

        :param bundle: The bundle to uninstall
        :raise BundleException: Invalid bundle
        """
        if bundle is None:
            # Do nothing
            return

        with self.__bundles_lock:
            # Stop the bundle first
            bundle.stop()

            bundle_id = bundle.get_bundle_id()
            if bundle_id not in self.__bundles:
                raise BundleException("Invalid bundle {0}".format(bundle))

            # Notify listeners
            self._dispatcher.fire_bundle_event(
                BundleEvent(BundleEvent.UNINSTALLED, bundle)
            )

            # Remove it from the dictionary
            del self.__bundles[bundle_id]

            # Remove it from the system => avoid unintended behaviors and
            # forces a complete module reload if it is re-installed
            name = bundle.get_symbolic_name()
            try:
                del sys.modules[name]
            except KeyError:
                # Ignore
                pass

            try:
                # Clear reference in parent
                parent, basename = name.rsplit(".", 1)
                if parent:
                    delattr(sys.modules[parent], basename)
            except (KeyError, AttributeError, ValueError):
                # Ignore errors
                pass

    def unregister_service(self, registration):
        # type: (ServiceRegistration) -> bool
        """
        Unregisters the given service

        :param registration: A ServiceRegistration to the service to unregister
        :raise BundleException: Invalid reference
        """
        # Get the Service Reference
        reference = registration.get_reference()

        # Remove the service from the registry
        svc_instance = self._registry.unregister(reference)

        # Keep a track of the unregistering reference
        self.__unregistering_services[reference] = svc_instance

        # Call the listeners
        event = ServiceEvent(ServiceEvent.UNREGISTERING, reference)
        self._dispatcher.fire_service_event(event)

        # Update the bundle registration information
        bundle = reference.get_bundle()
        bundle._unregistered_service(registration)

        # Remove the unregistering reference
        del self.__unregistering_services[reference]
        return True

    def _hide_bundle_services(self, bundle):
        # type: (Bundle) -> List[ServiceReference]
        """
        Hides the services of the given bundle in the service registry

        :param bundle: The bundle providing services
        :return: The references of the hidden services
        """
        return self._registry.hide_bundle_services(bundle)

    def _unget_used_services(self, bundle):
        # type: (Bundle) -> None
        """
        Cleans up all service usages of the given bundle

        :param bundle: Bundle to be cleaned up
        """
        self._registry.unget_used_services(bundle)

    def update(self):
        """
        Stops and starts the framework, if the framework is active.

        :raise BundleException: Something wrong occurred while stopping or
                                starting the framework.
        """
        with self._lock:
            if self._state == Bundle.ACTIVE:
                self.stop()
                self.start()

    def wait_for_stop(self, timeout=None):
        # type: (Optional[int]) -> bool
        """
        Waits for the framework to stop. Does nothing if the framework bundle
        is not in ACTIVE state.

        Uses a threading.Condition object

        :param timeout: The maximum time to wait (in seconds)
        :return: True if the framework has stopped, False if the timeout raised
        """
        if self._state != Bundle.ACTIVE:
            # Inactive framework, ignore the call
            return True

        self._fw_stop_event.wait(timeout)

        with self._lock:
            # If the timeout raised, we should be in another state
            return self._state == Bundle.RESOLVED


# ------------------------------------------------------------------------------


class ServiceObjects(object):
    """
    Allows multiple service objects for a service to be obtained.
    """

    def __init__(self, registry, bundle, svc_ref):
        # type: (ServiceRegistry, Bundle, ServiceReference) -> None
        """
        :param bundle: Bundle requesting the service
        :param svc_ref: Reference to the requested service
        """
        self.__registry = registry
        self.__bundle = bundle
        self.__reference = svc_ref

    def get_service(self):
        # type: () -> Any
        """
        Returns a service object for the associated service.
        """
        return self.__registry.get_service(self.__bundle, self.__reference)

    def get_service_reference(self):
        # type: () -> ServiceReference
        """
        Returns the ServiceReference for the service associated with this
        object.

        :return: The ServiceReference to the service associated to this object
        """
        return self.__reference

    def unget_service(self, service):
        # type: (Any) -> bool
        """
        Releases a service object for the associated service.

        :param service: An instance of a service returned by ``get_service()``
        :return: True if the bundle usage has been removed
        """
        return self.__registry.unget_service(
            self.__bundle, self.__reference, service
        )


class BundleContext(object):
    # pylint: disable=W0212
    """
    The bundle context is the link between a bundle and the framework.
    It is unique for a bundle and is created by the framework once the bundle
    is installed.
    """

    def __init__(self, framework, bundle):
        # type: (Framework, Bundle) -> None
        """
        :param framework: Hosting framework
        :param bundle: The associated bundle
        """
        self.__bundle = bundle
        self.__framework = framework

    def __str__(self):
        """
        String representation
        """
        return "BundleContext({0})".format(self.__bundle)

    def add_bundle_listener(self, listener):
        """
        Registers a bundle listener, which will be notified each time a bundle
        is installed, started, stopped or updated.

        The listener must be a callable accepting a single parameter:\

           * **event** -- The description of the event
             (a :class:`~pelix.internals.events.BundleEvent` object).

        :param listener: The bundle listener to register
        :return: True if the listener has been registered, False if it already
                 was
        """
        return self.__framework._dispatcher.add_bundle_listener(listener)

    def add_framework_stop_listener(self, listener):
        """
        Registers a listener that will be called back right before the
        framework stops

        The framework listener must have a method with the following prototype::

           def framework_stopping(self):
               '''
               No parameter given
               '''
               # ...

        :param listener: The framework stop listener
        :return: True if the listener has been registered
        """
        return self.__framework._dispatcher.add_framework_listener(listener)

    def add_service_listener(
        self, listener, ldap_filter=None, specification=None
    ):
        """
        Registers a service listener

        The service listener must have a method with the following prototype::

           def service_changed(self, event):
               '''
               Called by Pelix when some service properties changes

               event: A ServiceEvent object
               '''
               # ...

        :param bundle_context:  This bundle context
        :param listener: The listener to register
        :param ldap_filter: Filter that must match the service properties
                            (optional, None to accept all services)
        :param specification: The specification that must provide the service
                              (optional, None to accept all services)
        :return: True if the listener has been successfully registered
        """
        return self.__framework._dispatcher.add_service_listener(
            self, listener, specification, ldap_filter
        )

    def get_all_service_references(self, clazz, ldap_filter=None):
        """
        Returns an array of ServiceReference objects.
        The returned array of ServiceReference objects contains services that
        were registered under the specified class and match the specified
        filter expression.

        :param clazz: Class implemented by the service
        :param ldap_filter: Service filter
        :return: The sorted list of all matching service references, or None
        """
        return self.__framework.find_service_references(clazz, ldap_filter)

    def get_bundle(self, bundle_id=None):
        # type: (Union[Bundle, int]) -> Bundle
        """
        Retrieves the :class:`~pelix.framework.Bundle` object for the bundle
        matching the given ID (int). If no ID is given (None), the bundle
        associated to this context is returned.

        :param bundle_id: A bundle ID (optional)
        :return: The requested :class:`~pelix.framework.Bundle` object
        :raise BundleException: The given ID doesn't exist or is invalid
        """
        if bundle_id is None:
            # Current bundle
            return self.__bundle
        elif isinstance(bundle_id, Bundle):
            # Got a bundle (compatibility with older install_bundle())
            bundle_id = bundle_id.get_bundle_id()

        return self.__framework.get_bundle_by_id(bundle_id)

    def get_bundles(self):
        # type: () -> List[Bundle]
        """
        Returns the list of all installed bundles

        :return: A list of :class:`~pelix.framework.Bundle` objects
        """
        return self.__framework.get_bundles()

    def get_framework(self):
        # type: () -> Framework
        """
        Returns the :class:`~pelix.FRAMEWORK.Framework` that created this
        bundle context

        :return: The :class:`~pelix.framework.Framework` object
        """
        return self.__framework

    def get_property(self, name):
        # type: (str) -> object
        """
        Returns the value of a property of the framework, else returns the OS
        environment value.

        :param name: A property name
        """
        return self.__framework.get_property(name)

    def get_service(self, reference):
        # type: (ServiceReference) -> Any
        """
        Returns the service described with the given reference

        :param reference: A ServiceReference object
        :return: The service object itself
        """
        return self.__framework.get_service(self.__bundle, reference)

    def get_service_objects(self, reference):
        # type: (ServiceReference) -> ServiceObjects
        """
        Returns the ServiceObjects object for the service referenced by the
        specified ServiceReference object.

        :param reference: Reference to a prototype service factory
        :return: An intermediate object to get more instances of a service
        """
        return self.__framework._get_service_objects(self.__bundle, reference)

    def get_service_reference(self, clazz, ldap_filter=None):
        # type: (Optional[str], Optional[str]) -> Optional[ServiceReference]
        """
        Returns a ServiceReference object for a service that implements and
        was registered under the specified class

        :param clazz: The class name with which the service was registered.
        :param ldap_filter: A filter on service properties
        :return: A service reference, None if not found
        """
        result = self.__framework.find_service_references(
            clazz, ldap_filter, True
        )
        try:
            return result[0]
        except TypeError:
            return None

    def get_service_references(self, clazz, ldap_filter=None):
        # type: (Optional[str], Optional[str]) -> Optional[List[ServiceReference]]
        """
        Returns the service references for services that were registered under
        the specified class by this bundle and matching the given filter

        :param clazz: The class name with which the service was registered.
        :param ldap_filter: A filter on service properties
        :return: The list of references to the services registered by the
                 calling bundle and matching the filters.
        """
        refs = self.__framework.find_service_references(clazz, ldap_filter)
        if refs:
            for ref in refs:
                if ref.get_bundle() is not self.__bundle:
                    refs.remove(ref)

        return refs

    def install_bundle(self, name, path=None):
        # type: (str, str) -> Bundle
        """
        Installs the bundle (module) with the given name.

        If a path is given, it is inserted in first place in the Python loading
        path (``sys.path``). All modules loaded alongside this bundle, *i.e.*
        by this bundle or its dependencies, will be looked after in this path
        in priority.

        .. note::
            Before Pelix 0.5.0, this method returned the ID of the installed
            bundle, instead of the Bundle object.

        .. warning::
            The behavior of the loading process is subject to changes, as it
            does not allow to safely run multiple frameworks in the same Python
            interpreter, as they might share global module values.

        :param name: The name of the bundle to install
        :param path: Preferred path to load the module (optional)
        :return: The :class:`~pelix.framework.Bundle` object of the installed
                 bundle
        :raise BundleException: Error importing the module or one of its
                                dependencies
        """
        return self.__framework.install_bundle(name, path)

    def install_package(self, path, recursive=False):
        # type: (str, bool) -> tuple
        """
        Installs all the modules found in the given package (directory).
        It is a utility method working like
        :meth:`~pelix.framework.BundleContext.install_visiting`, with a visitor
        accepting every module found.

        :param path: Path of the package (folder)
        :param recursive: If True, installs the modules found in sub-directories
        :return: A 2-tuple, with the list of installed bundles
                 (:class:`~pelix.framework.Bundle`) and the list of the names
                 of the modules which import failed.
        :raise ValueError: The given path is invalid
        """
        return self.__framework.install_package(path, recursive)

    def install_visiting(self, path, visitor):
        """
        Looks for modules in the given path and installs those accepted by the
        given visitor.

        The visitor must be a callable accepting 3 parameters:\

           * **fullname** -- The full name of the module
           * **is_package** -- If True, the module is a package
           * **module_path** -- The path to the module file

        :param path: Root search path (folder)
        :param visitor: The visiting callable
        :return: A 2-tuple, with the list of installed bundles
                 (:class:`~pelix.framework.Bundle`) and the list of the names
                 of the modules which import failed.
        :raise ValueError: Invalid path or visitor
        """
        return self.__framework.install_visiting(path, visitor)

    def register_service(
        self,
        clazz,
        service,
        properties,
        send_event=True,
        factory=False,
        prototype=False,
    ):
        # type: (Union[List[Any], type, str], object, dict, bool, bool, bool) -> ServiceRegistration
        """
        Registers a service

        :param clazz: Class or Classes (list) implemented by this service
        :param service: The service instance
        :param properties: The services properties (dictionary)
        :param send_event: If not, doesn't trigger a service registered event
        :param factory: If True, the given service is a service factory
        :param prototype: If True, the given service is a prototype service
                          factory (the factory argument is considered True)
        :return: A ServiceRegistration object
        :raise BundleException: An error occurred while registering the service
        """
        return self.__framework.register_service(
            self.__bundle,
            clazz,
            service,
            properties,
            send_event,
            factory,
            prototype,
        )

    def remove_bundle_listener(self, listener):
        """
        Unregisters the given bundle listener

        :param listener: The bundle listener to remove
        :return: True if the listener has been unregistered,
                 False if it wasn't registered
        """
        return self.__framework._dispatcher.remove_bundle_listener(listener)

    def remove_framework_stop_listener(self, listener):
        """
        Unregisters a framework stop listener

        :param listener: The framework stop listener
        :return: True if the listener has been unregistered
        """
        return self.__framework._dispatcher.remove_framework_listener(listener)

    def remove_service_listener(self, listener):
        """
        Unregisters a service listener

        :param listener: The service listener
        :return: True if the listener has been unregistered
        """
        return self.__framework._dispatcher.remove_service_listener(listener)

    def unget_service(self, reference):
        # type: (ServiceReference) -> bool
        """
        Disables a reference to the service

        :return: True if the bundle was using this reference, else False
        """
        # Lose the dependency
        return self.__framework._registry.unget_service(
            self.__bundle, reference
        )


# ------------------------------------------------------------------------------


class FrameworkFactory(object):
    """
    A framework factory
    """

    __singleton = None  # type: Framework
    """ The framework singleton """

    @classmethod
    def get_framework(cls, properties=None):
        # type: (Optional[dict]) -> Framework
        """
        If it doesn't exist yet, creates a framework with the given properties,
        else returns the current framework instance.

        :return: A Pelix instance
        """
        if cls.__singleton is None:
            # Normalize sys.path
            normalize_path()
            cls.__singleton = Framework(properties)

        return cls.__singleton

    @classmethod
    def is_framework_running(cls, framework=None):
        # type: (Optional[Framework]) -> bool
        """
        Tests if the given framework has been constructed and not deleted.
        If *framework* is None, then the methods returns if at least one
        framework is running.

        :param framework: The framework instance to be tested
        :return: True if the framework is running
        """
        if framework is None:
            return cls.__singleton is not None

        return cls.__singleton == framework

    @classmethod
    def delete_framework(cls, framework=None):
        # type: (Optional[Framework]) -> bool
        # pylint: disable=W0212
        """
        Removes the framework singleton

        :return: True on success, else False
        """
        if framework is None:
            framework = cls.__singleton

        if framework is cls.__singleton:
            # Stop the framework
            try:
                framework.stop()
            except:
                _logger.exception("Error stopping the framework")

            # Uninstall its bundles
            bundles = framework.get_bundles()
            for bundle in bundles:
                try:
                    bundle.uninstall()
                except:
                    _logger.exception(
                        "Error uninstalling bundle %s",
                        bundle.get_symbolic_name(),
                    )

            # Clear the event dispatcher
            framework._dispatcher.clear()

            # Clear the singleton
            cls.__singleton = None
            return True

        return False


# ------------------------------------------------------------------------------


def create_framework(
    bundles,
    properties=None,
    auto_start=False,
    wait_for_stop=False,
    auto_delete=False,
):
    # type: (Union[list, tuple], dict, bool, bool, bool) -> Framework
    """
    Creates a Pelix framework, installs the given bundles and returns its
    instance reference.
    If *auto_start* is True, the framework will be started once all bundles
    will have been installed
    If *wait_for_stop* is True, the method will return only when the framework
    will have stopped. This requires *auto_start* to be True.
    If *auto_delete* is True, the framework will be deleted once it has
    stopped, and the method will return None.
    This requires *wait_for_stop* and *auto_start* to be True.

    :param bundles: Bundles to initially install (shouldn't be empty if
                    *wait_for_stop* is True)
    :param properties: Optional framework properties
    :param auto_start: If True, the framework will be started immediately
    :param wait_for_stop: If True, the method will return only when the
                          framework will have stopped
    :param auto_delete: If True, deletes the framework once it stopped.
    :return: The framework instance
    :raise ValueError: Only one framework can run at a time
    """
    # Test if a framework already exists
    if FrameworkFactory.is_framework_running(None):
        raise ValueError("A framework is already running")

    # Create the framework
    framework = FrameworkFactory.get_framework(properties)

    # Install bundles
    context = framework.get_bundle_context()
    for bundle in bundles:
        context.install_bundle(bundle)

    if auto_start:
        # Automatically start the framework
        framework.start()

        if wait_for_stop:
            # Wait for the framework to stop
            try:
                framework.wait_for_stop(None)
            except KeyboardInterrupt:
                # Stop keyboard interruptions
                if framework.get_state() == Bundle.ACTIVE:
                    framework.stop()

            if auto_delete:
                # Delete the framework
                FrameworkFactory.delete_framework(framework)
                framework = None

    return framework


def _package_exists(path):
    # type: (str) -> bool
    """
    Checks if the given Python path matches a valid file or a valid container
    file

    :param path: A Python path
    :return: True if the module or its container exists
    """
    while path:
        if os.path.exists(path):
            return True
        else:
            path = os.path.dirname(path)

    return False


def normalize_path():
    """
    Normalizes sys.path to avoid the use of relative folders
    """
    # Normalize Python paths
    whole_path = [
        os.path.abspath(path) for path in sys.path if os.path.exists(path)
    ]

    # Keep the "dynamic" current folder indicator and add the "static"
    # current path
    # Use an OrderedDict to have a faster lookup (path not in whole_set)
    whole_set = collections.OrderedDict((("", 1), (os.getcwd(), 1)))

    # Add original path entries
    for path in whole_path:
        if path not in whole_set:
            whole_set[path] = 1

    # Set the new content of sys.path (still ordered thanks to OrderedDict)
    sys.path = list(whole_set)

    # Normalize paths in loaded modules
    for module_ in sys.modules.values():
        try:
            if module_.__path__ is not None:
                # Seems that (some?) DLL-based modules don't have a __path__
                # but their __file__ is already absolute
                module_.__path__ = [
                    os.path.abspath(path)
                    for path in module_.__path__
                    if _package_exists(path)
                ]
        except AttributeError:
            # builtin modules don't have a __path__
            pass
        except ImportError:
            pass



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Core iPOPO implementation

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import copy
import inspect
import logging
import sys
import threading

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Dict, List, Optional, Set, Tuple
    from pelix.framework import BundleContext, ServiceReference
except ImportError:
    pass

# Pelix
from pelix.constants import SERVICE_ID, BundleActivator
from pelix.framework import Bundle, BundleException
from pelix.internals.events import BundleEvent, ServiceEvent
from pelix.utilities import add_listener, remove_listener, is_string

# iPOPO constants
import pelix.ipopo.constants as constants
import pelix.ipopo.handlers.constants as handlers_const

# iPOPO beans
from pelix.ipopo.contexts import FactoryContext, ComponentContext
from pelix.ipopo.instance import StoredInstance

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

# Prepare the module logger
_logger = logging.getLogger("ipopo.core")

# Built-in handlers, automatically installed
BUILTIN_HANDLERS = (
    "pelix.ipopo.handlers.properties",
    "pelix.ipopo.handlers.provides",
    "pelix.ipopo.handlers.requires",
    "pelix.ipopo.handlers.requiresbest",
    "pelix.ipopo.handlers.requiresbroadcast",
    "pelix.ipopo.handlers.requiresmap",
    "pelix.ipopo.handlers.requiresvarfilter",
    "pelix.ipopo.handlers.temporal",
)

# ------------------------------------------------------------------------------


def _set_factory_context(factory_class, bundle_context):
    # type: (type, Optional[BundleContext]) -> Optional[FactoryContext]
    """
    Transforms the context data dictionary into its FactoryContext object form.

    :param factory_class: A manipulated class
    :param bundle_context: The class bundle context
    :return: The factory context, None on error
    """
    try:
        # Try to get the factory context (built using decorators)
        context = getattr(factory_class, constants.IPOPO_FACTORY_CONTEXT)
    except AttributeError:
        # The class has not been manipulated, or too badly
        return None

    if not context.completed:
        # Partial context (class not manipulated)
        return None

    # Associate the factory to the bundle context
    context.set_bundle_context(bundle_context)
    return context


def _load_bundle_factories(bundle):
    # type: (Bundle) -> List[Tuple[FactoryContext, type]]
    """
    Retrieves a list of pairs (FactoryContext, factory class) with all
    readable manipulated classes found in the bundle.

    :param bundle: A Bundle object
    :return: The list of factories loaded from the bundle
    """
    result = []

    # Get the Python module
    module_ = bundle.get_module()

    # Get the bundle context
    bundle_context = bundle.get_bundle_context()

    for name in dir(module_):
        try:
            # Get the module member
            factory_class = getattr(module_, name)
            if not inspect.isclass(factory_class):
                continue

            # Check if it is a class
            if sys.modules[factory_class.__module__] is not module_:
                # Only keep classes from this module
                continue
        except (AttributeError, KeyError):
            # getattr() didn't work or __module__ is not a member of the class,
            # or the module is not known by the interpreter
            continue

        context = _set_factory_context(factory_class, bundle_context)
        if context is None:
            # Error setting up the factory context
            continue

        result.append((context, factory_class))

    return result


# ------------------------------------------------------------------------------


class _IPopoService(object):
    """
    The iPOPO registry and service.

    This service is registered automatically and **must not** be started
    manually.
    """

    def __init__(self, bundle_context):
        # type: (BundleContext) -> None
        """
        :param bundle_context: The iPOPO bundle context
        """
        # Store the bundle context
        self.__context = bundle_context

        # Factories registry : name -> factory class
        self.__factories = {}  # type: Dict[str, type]

        # Instances registry : name -> StoredInstance object
        self.__instances = {}  # type: Dict[str, StoredInstance]

        # Event listeners
        self.__listeners = []  # type: List[Any]

        # Auto-restarted components (Bundle -> [(factory, name, properties)]
        self.__auto_restart = {}  # type: Dict[Bundle, List[Tuple[str, str, Dict]]]

        # Service state
        self.running = False

        # Registries locks
        self.__factories_lock = threading.RLock()
        self.__instances_lock = threading.RLock()
        self.__listeners_lock = threading.RLock()
        self.__handlers_lock = threading.RLock()

        # Handlers factories
        self._handlers_refs = set()  # type: Set[ServiceReference]
        self._handlers = {}  # type: Dict[str, Any]

        # Instances waiting for a handler: Name -> (ComponentContext, instance)
        self.__waiting_handlers = {}  # type: Dict[str, Tuple[ComponentContext, Any]]

        # Register the service listener
        bundle_context.add_service_listener(
            self, None, handlers_const.SERVICE_IPOPO_HANDLER_FACTORY
        )
        self.__find_handler_factories()

    def __find_handler_factories(self):
        """
        Finds all registered handler factories and stores them
        """
        # Get the references
        svc_refs = self.__context.get_all_service_references(
            handlers_const.SERVICE_IPOPO_HANDLER_FACTORY
        )
        if svc_refs:
            for svc_ref in svc_refs:
                # Store each handler factory
                self.__add_handler_factory(svc_ref)

    def __add_handler_factory(self, svc_ref):
        # type: (ServiceReference) -> None
        """
        Stores a new handler factory

        :param svc_ref: ServiceReference of the new handler factory
        """
        with self.__handlers_lock:
            # Get the handler ID
            handler_id = svc_ref.get_property(handlers_const.PROP_HANDLER_ID)
            if handler_id in self._handlers:
                # Duplicated ID
                _logger.warning("Already registered handler ID: %s", handler_id)
            else:
                # Store the service
                self._handlers_refs.add(svc_ref)
                self._handlers[handler_id] = self.__context.get_service(svc_ref)

                # Try to instantiate waiting components
                succeeded = set()
                for (
                    name,
                    (context, instance),
                ) in self.__waiting_handlers.items():
                    if self.__try_instantiate(context, instance):
                        succeeded.add(name)

                # Remove instantiated component from the waiting list
                for name in succeeded:
                    del self.__waiting_handlers[name]

    def __remove_handler_factory(self, svc_ref):
        # type: (ServiceReference) -> None
        """
        Removes an handler factory

        :param svc_ref: ServiceReference of the handler factory to remove
        """
        with self.__handlers_lock:
            # Get the handler ID
            handler_id = svc_ref.get_property(handlers_const.PROP_HANDLER_ID)

            # Check if this is the handler we use
            if svc_ref not in self._handlers_refs:
                return

            # Clean up
            self.__context.unget_service(svc_ref)
            self._handlers_refs.remove(svc_ref)
            del self._handlers[handler_id]

            # List the components using this handler
            to_stop = set()  # type: Set[StoredInstance]
            for factory_name in self.__factories:
                _, factory_context = self.__get_factory_with_context(
                    factory_name
                )
                if handler_id in factory_context.get_handlers_ids():
                    to_stop.update(self.__get_stored_instances(factory_name))

            with self.__instances_lock:
                for stored_instance in to_stop:
                    # Extract information
                    context = stored_instance.context
                    name = context.name
                    instance = stored_instance.instance

                    # Clean up the stored instance (iPOPO side)
                    del self.__instances[name]
                    stored_instance.kill()

                    # Add the component to the waiting queue
                    self.__waiting_handlers[name] = (context, instance)

            # Try to find a new handler factory
            new_ref = self.__context.get_service_reference(
                handlers_const.SERVICE_IPOPO_HANDLER_FACTORY,
                "({0}={1})".format(handlers_const.PROP_HANDLER_ID, handler_id),
            )
            if new_ref is not None:
                self.__add_handler_factory(new_ref)

    def __get_factory_with_context(self, factory_name):
        # type: (str) -> Tuple[type, FactoryContext]
        """
        Retrieves the factory registered with the given and its factory context

        :param factory_name: The name of the factory
        :return: A (factory, context) tuple
        :raise TypeError: Unknown factory, or factory not manipulated
        """
        factory = self.__factories.get(factory_name)
        if factory is None:
            raise TypeError("Unknown factory '{0}'".format(factory_name))

        # Get the factory context
        factory_context = getattr(
            factory, constants.IPOPO_FACTORY_CONTEXT, None
        )
        if factory_context is None:
            raise TypeError(
                "Factory context missing in '{0}'".format(factory_name)
            )

        return factory, factory_context

    def __get_handler_factories(self, handlers_ids):
        # type: (list) -> set
        """
        Returns the list of Handler Factories for the given Handlers IDs.
        Raises a KeyError exception is a handler factory is missing.

        :param handlers_ids: List of handlers IDs
        :raise KeyError: A handler is missing
        """
        # Look for the required handlers
        return {self._handlers[handler_id] for handler_id in handlers_ids}

    def __get_stored_instances(self, factory_name):
        # type: (str) -> List[StoredInstance]
        """
        Retrieves the list of all stored instances objects corresponding to
        the given factory name

        :param factory_name: A factory name
        :return: All components instantiated from the given factory
        """
        with self.__instances_lock:
            return [
                stored_instance
                for stored_instance in self.__instances.values()
                if stored_instance.factory_name == factory_name
            ]

    def __try_instantiate(self, component_context, instance):
        # type: (ComponentContext, object) -> bool
        """
        Instantiates a component, if all of its handlers are there. Returns
        False if a handler is missing.

        :param component_context: A ComponentContext bean
        :param instance: The component instance
        :return: True if the component has started,
                 False if a handler is missing
        """
        with self.__instances_lock:
            # Extract information about the component
            factory_context = component_context.factory_context
            handlers_ids = factory_context.get_handlers_ids()
            name = component_context.name
            factory_name = factory_context.name

            try:
                # Get handlers
                handler_factories = self.__get_handler_factories(handlers_ids)
            except KeyError:
                # A handler is missing, stop here
                return False

            # Instantiate the handlers
            all_handlers = set()  # type: Set[Any]
            for handler_factory in handler_factories:
                handlers = handler_factory.get_handlers(
                    component_context, instance
                )
                if handlers:
                    all_handlers.update(handlers)

            # Prepare the stored instance
            stored_instance = StoredInstance(
                self, component_context, instance, all_handlers
            )

            # Manipulate the properties
            for handler in all_handlers:
                handler.manipulate(stored_instance, instance)

            # Store the instance
            self.__instances[name] = stored_instance

        # Start the manager
        stored_instance.start()

        # Notify listeners now that every thing is ready to run
        self._fire_ipopo_event(
            constants.IPopoEvent.INSTANTIATED, factory_name, name
        )

        # Try to validate it
        stored_instance.update_bindings()
        stored_instance.check_lifecycle()
        return True

    def _autorestart_store_components(self, bundle):
        # type: (Bundle) -> None
        """
        Stores the components of the given bundle with the auto-restart
        property

        :param bundle: A Bundle object
        """
        with self.__instances_lock:
            # Prepare the list of components
            store = self.__auto_restart.setdefault(bundle, [])
            for stored_instance in self.__instances.values():
                # Get the factory name
                factory = stored_instance.factory_name
                if self.get_factory_bundle(factory) is bundle:
                    # Factory from this bundle
                    # Test component properties
                    properties = stored_instance.context.properties
                    if properties.get(constants.IPOPO_AUTO_RESTART):
                        # Auto-restart property found
                        store.append(
                            (factory, stored_instance.name, properties)
                        )

    def _autorestart_components(self, bundle):
        # type: (Bundle) -> None
        """
        Restart the components of the given bundle

        :param bundle: A Bundle object
        """
        with self.__instances_lock:
            instances = self.__auto_restart.get(bundle)
            if not instances:
                # Nothing to do
                return

            for factory, name, properties in instances:
                try:
                    # Instantiate the given component
                    self.instantiate(factory, name, properties)
                except Exception as ex:
                    # Log error, but continue to work
                    _logger.exception(
                        "Error restarting component '%s' ('%s') "
                        "from bundle %s (%d): %s",
                        name,
                        factory,
                        bundle.get_symbolic_name(),
                        bundle.get_bundle_id(),
                        ex,
                    )

    def _autorestart_clear_components(self, bundle):
        # type: (Bundle) -> None
        """
        Clear the list of auto-restart components of the given bundle

        :param bundle: A Bundle object
        """
        with self.__instances_lock:
            # Simply delete the entry, if any
            try:
                del self.__auto_restart[bundle]
            except KeyError:
                pass

    def _fire_ipopo_event(self, kind, factory_name, instance_name=None):
        # type: (int, str, Optional[str]) -> None
        """
        Triggers an iPOPO event

        :param kind: Kind of event
        :param factory_name: Name of the factory associated to the event
        :param instance_name: Name of the component instance associated to the
                              event
        """
        with self.__listeners_lock:
            # Use a copy of the list of listeners
            listeners = self.__listeners[:]

        for listener in listeners:
            try:
                listener.handle_ipopo_event(
                    constants.IPopoEvent(kind, factory_name, instance_name)
                )
            except:
                _logger.exception("Error calling an iPOPO event handler")

    def _prepare_instance_properties(self, properties, factory_properties):
        # type: (dict, dict) -> dict
        """
        Prepares the properties of a component instance, based on its
        configuration, factory and framework properties

        :param properties: Component instance properties
        :param factory_properties: Component factory "default" properties
        :return: The merged properties
        """
        # Normalize given properties
        if properties is None or not isinstance(properties, dict):
            properties = {}

        # Use framework properties to fill missing ones
        framework = self.__context.get_framework()
        for property_name in factory_properties:
            if property_name not in properties:
                # Missing property
                value = framework.get_property(property_name)
                if value is not None:
                    # Set the property value
                    properties[property_name] = value

        return properties

    def _register_bundle_factories(self, bundle):
        # type: (Bundle) -> None
        """
        Registers all factories found in the given bundle

        :param bundle: A bundle
        """
        # Load the bundle factories
        factories = _load_bundle_factories(bundle)

        for context, factory_class in factories:
            try:
                # Register each found factory
                self._register_factory(context.name, factory_class, False)
            except ValueError as ex:
                # Already known factory
                _logger.error(
                    "Cannot register factory '%s' of bundle %d (%s): %s",
                    context.name,
                    bundle.get_bundle_id(),
                    bundle.get_symbolic_name(),
                    ex,
                )
                _logger.error(
                    "class: %s -- module: %s",
                    factory_class,
                    factory_class.__module__,
                )
            else:
                # Instantiate components
                for name, properties in context.get_instances().items():
                    self.instantiate(context.name, name, properties)

    def _register_factory(self, factory_name, factory, override):
        # type: (str, type, bool) -> None
        """
        Registers a component factory

        :param factory_name: The name of the factory
        :param factory: The factory class object
        :param override: If true, previous factory is overridden, else an
                         exception is risen if a previous factory with that
                         name already exists
        :raise ValueError: The factory name already exists or is invalid
        :raise TypeError: Invalid factory type
        """
        if not factory_name or not is_string(factory_name):
            raise ValueError("A factory name must be a non-empty string")

        if not inspect.isclass(factory):
            raise TypeError(
                "Invalid factory class '{0}'".format(type(factory).__name__)
            )

        with self.__factories_lock:
            if factory_name in self.__factories:
                if override:
                    _logger.info("Overriding factory '%s'", factory_name)
                else:
                    raise ValueError(
                        "'{0}' factory already exist".format(factory_name)
                    )

            self.__factories[factory_name] = factory

            # Trigger an event
            self._fire_ipopo_event(
                constants.IPopoEvent.REGISTERED, factory_name
            )

    def _unregister_all_factories(self):
        """
        Unregisters all factories. This method should be called only after the
        iPOPO service has been unregistered (that's why it's not locked)
        """
        factories = list(self.__factories.keys())
        for factory_name in factories:
            self.unregister_factory(factory_name)

    def _unregister_bundle_factories(self, bundle):
        # type: (Bundle) -> None
        """
        Unregisters all factories of the given bundle

        :param bundle: A bundle
        """
        with self.__factories_lock:
            # Find out which factories must be removed
            to_remove = [
                factory_name
                for factory_name in self.__factories
                if self.get_factory_bundle(factory_name) is bundle
            ]

            # Remove all of them
            for factory_name in to_remove:
                try:
                    self.unregister_factory(factory_name)
                except ValueError as ex:
                    _logger.warning(
                        "Error unregistering factory '%s': %s", factory_name, ex
                    )

    def _stop(self):
        """
        iPOPO is stopping: clean everything up
        """
        # Running flag down
        self.running = False

        # Unregister the service listener
        self.__context.remove_service_listener(self)

        # Clean up handler factories usages
        with self.__instances_lock:
            for svc_ref in self._handlers_refs:
                self.__context.unget_service(svc_ref)

            self._handlers.clear()
            self._handlers_refs.clear()

    def framework_stopping(self):
        """
        Called by the framework when it is about to stop
        """
        self._stop()

    def bundle_changed(self, event):
        # type: (BundleEvent) -> None
        """
        A bundle event has been triggered

        :param event: The bundle event
        """
        kind = event.get_kind()
        bundle = event.get_bundle()

        if kind == BundleEvent.STOPPING_PRECLEAN:
            # A bundle is gone, remove its factories after the deactivator has
            # been called. That way, the deactivator can kill manually started
            # components.
            self._unregister_bundle_factories(bundle)

        elif kind == BundleEvent.STARTED:
            # A bundle is staring, register its factories before its activator
            # is called. That way, the activator can use the registered
            # factories.
            self._register_bundle_factories(bundle)

        elif kind == BundleEvent.UPDATE_BEGIN:
            # A bundle will be updated, store its auto-restart component
            self._autorestart_store_components(bundle)

        elif kind == BundleEvent.UPDATED:
            # Update has finished, restart stored components
            self._autorestart_components(bundle)
            self._autorestart_clear_components(bundle)

        elif kind == BundleEvent.UPDATE_FAILED:
            # Update failed, clean the stored components
            self._autorestart_clear_components(bundle)

    def service_changed(self, event):
        # type: (ServiceEvent) -> None
        """
        Called when a handler factory service is un/registered
        """
        # Call sub-methods
        kind = event.get_kind()
        svc_ref = event.get_service_reference()

        if kind == ServiceEvent.REGISTERED:
            # Service coming
            with self.__instances_lock:
                self.__add_handler_factory(svc_ref)

        elif kind == ServiceEvent.UNREGISTERING:
            # Service gone
            with self.__instances_lock:
                self.__remove_handler_factory(svc_ref)

    def instantiate(self, factory_name, name, properties=None):
        # type: (str, str, dict) -> Any
        """
        Instantiates a component from the given factory, with the given name

        :param factory_name: Name of the component factory
        :param name: Name of the instance to be started
        :param properties: Initial properties of the component instance
        :return: The component instance
        :raise TypeError: The given factory is unknown
        :raise ValueError: The given name or factory name is invalid, or an
                           instance with the given name already exists
        :raise Exception: Something wrong occurred in the factory
        """
        # Test parameters
        if not factory_name or not is_string(factory_name):
            raise ValueError("Invalid factory name")

        if not name or not is_string(name):
            raise ValueError("Invalid component name")

        if not self.running:
            # Stop working if the framework is stopping
            raise ValueError("Framework is stopping")

        with self.__instances_lock:
            if name in self.__instances or name in self.__waiting_handlers:
                raise ValueError(
                    "'{0}' is an already running instance name".format(name)
                )

            with self.__factories_lock:
                # Can raise a TypeError exception
                factory, factory_context = self.__get_factory_with_context(
                    factory_name
                )

                # Check if the factory is singleton and if a component is
                # already started
                if (
                    factory_context.is_singleton
                    and factory_context.is_singleton_active
                ):
                    raise ValueError(
                        "{0} is a singleton: {1} can't be "
                        "instantiated.".format(factory_name, name)
                    )

                # Create component instance
                try:
                    instance = factory()
                except Exception:
                    _logger.exception(
                        "Error creating the instance '%s' from factory '%s'",
                        name,
                        factory_name,
                    )
                    raise TypeError(
                        "Factory '{0}' failed to create '{1}'".format(
                            factory_name, name
                        )
                    )

                # Instantiation succeeded: update singleton status
                if factory_context.is_singleton:
                    factory_context.is_singleton_active = True

            # Normalize the given properties
            properties = self._prepare_instance_properties(
                properties, factory_context.properties
            )

            # Set up the component instance context
            component_context = ComponentContext(
                factory_context, name, properties
            )

            # Try to instantiate the component immediately
            if not self.__try_instantiate(component_context, instance):
                # A handler is missing, put the component in the queue
                self.__waiting_handlers[name] = (component_context, instance)

        return instance

    def retry_erroneous(self, name, properties_update=None):
        # type: (str, dict) -> int
        """
        Removes the ERRONEOUS state of the given component, and retries a
        validation

        :param name: Name of the component to retry
        :param properties_update: A dictionary to update the initial properties
                                  of the component
        :return: The new state of the component
        :raise ValueError: Invalid component name
        """
        with self.__instances_lock:
            try:
                stored_instance = self.__instances[name]
            except KeyError:
                raise ValueError(
                    "Unknown component instance '{0}'".format(name)
                )
            else:
                return stored_instance.retry_erroneous(properties_update)

    def invalidate(self, name):
        # type: (str) -> None
        """
        Invalidates the given component

        :param name: Name of the component to invalidate
        :raise ValueError: Invalid component name
        """
        with self.__instances_lock:
            try:
                stored_instance = self.__instances[name]
            except KeyError:
                raise ValueError(
                    "Unknown component instance '{0}'".format(name)
                )
            else:
                # Call back the component during the invalidation
                stored_instance.invalidate(True)

    def is_registered_factory(self, name):
        # type: (str) -> bool
        """
        Tests if the given name is in the factory registry

        :param name: A factory name to be tested
        """
        with self.__factories_lock:
            return name in self.__factories

    def is_registered_instance(self, name):
        # type: (str) -> bool
        """
        Tests if the given name is in the instance registry or in the waiting
        queue

        :param name: A component name to be tested
        """
        with self.__instances_lock:
            return name in self.__instances

    def kill(self, name):
        # type: (str) -> None
        """
        Kills the given component

        :param name: Name of the component to kill
        :raise ValueError: Invalid component name
        """
        if not name:
            raise ValueError("Name can't be None or empty")

        with self.__instances_lock:
            try:
                # Running instance
                stored_instance = self.__instances.pop(name)

                # Store the reference to the factory context
                factory_context = stored_instance.context.factory_context

                # Kill it
                stored_instance.kill()

                # Update the singleton state flag
                factory_context.is_singleton_active = False
            except KeyError:
                # Queued instance
                try:
                    # Extract the component context
                    context, _ = self.__waiting_handlers.pop(name)

                    # Update the singleton state flag
                    context.factory_context.is_singleton_active = False
                except KeyError:
                    raise ValueError(
                        "Unknown component instance '{0}'".format(name)
                    )

    def register_factory(self, bundle_context, factory):
        # type: (BundleContext, type) -> bool
        """
        Registers a manually created factory, using decorators programmatically

        :param bundle_context: The factory bundle context
        :param factory: A manipulated class
        :return: True if the factory has been registered
        :raise ValueError: Invalid parameter, or factory already registered
        :raise TypeError: Invalid factory type (not a manipulated class)
        """
        if factory is None or bundle_context is None:
            # Invalid parameter, to nothing
            raise ValueError("Invalid parameter")

        context = _set_factory_context(factory, bundle_context)
        if not context:
            raise TypeError("Not a manipulated class (no context found)")

        self._register_factory(context.name, factory, False)
        return True

    def unregister_factory(self, factory_name):
        # type: (str) -> bool
        """
        Unregisters the given component factory

        :param factory_name: Name of the factory to unregister
        :return: True the factory has been removed, False if the factory is
                 unknown
        """
        if not factory_name or not is_string(factory_name):
            # Invalid name
            return False

        with self.__factories_lock:
            try:
                # Remove the factory from the registry
                factory_class = self.__factories.pop(factory_name)
            except KeyError:
                # Unknown factory
                return False

            # Trigger an event
            self._fire_ipopo_event(
                constants.IPopoEvent.UNREGISTERED, factory_name
            )

            # Invalidate and delete all components of this factory
            with self.__instances_lock:
                # Compute the list of __instances to remove
                to_remove = self.__get_stored_instances(factory_name)

                # Remove instances from the registry: avoids dependencies \
                # update to link against a component from this factory again.
                for instance in to_remove:
                    try:
                        # Kill the instance
                        self.kill(instance.name)
                    except ValueError:
                        # Unknown instance: already killed by the invalidation
                        # callback of a component killed in this loop
                        # => ignore
                        pass

                # Remove waiting component
                names = [
                    name
                    for name, (context, _) in self.__waiting_handlers.items()
                    if context.factory_context.name == factory_name
                ]
                for name in names:
                    del self.__waiting_handlers[name]

            # Clear the bundle context of the factory
            _set_factory_context(factory_class, None)

        return True

    def add_listener(self, listener):
        """
        Register an iPOPO event listener.

        The event listener must have a method with the following prototype::

           def handle_ipopo_event(self, event):
               '''
               event: A IPopoEvent object
               '''
               # ...

        :param listener: The listener to register
        :return: True if the listener has been added to the registry
        """
        with self.__listeners_lock:
            return add_listener(self.__listeners, listener)

    def remove_listener(self, listener):
        """
        Unregister an iPOPO event listener.

        :param listener: The listener to register
        :return: True if the listener has been removed from the registry
        """
        with self.__listeners_lock:
            return remove_listener(self.__listeners, listener)

    def get_instances(self):
        # type: () -> List[Tuple[str, str, int]]
        """
        Retrieves the list of the currently registered component instances

        :return: A list of (name, factory name, state) tuples.
        """
        with self.__instances_lock:
            return sorted(
                (name, stored_instance.factory_name, stored_instance.state)
                for name, stored_instance in self.__instances.items()
            )

    def get_instance(self, name):
        # type: (str) -> Any
        """
        Returns the instance of the component with the given name

        :param name: A component name
        :return: The component instance
        :raise KeyError: Unknown instance
        """
        return self.__instances[name].instance

    def get_waiting_components(self):
        # type: () -> List[Tuple[str, str, Set[str]]]
        """
        Returns the list of the instances waiting for their handlers

        :return: A list of (name, factory name, missing handlers) tuples
        """
        with self.__instances_lock:
            result = []
            for name, (context, _) in self.__waiting_handlers.items():
                # Compute missing handlers
                missing = set(context.factory_context.get_handlers_ids())
                missing.difference_update(self._handlers.keys())

                result.append((name, context.factory_context.name, missing))

            result.sort()
            return result

    def get_instance_details(self, name):
        # type: (str) -> Dict[str, Any]
        """
        Retrieves a snapshot of the given component instance.
        The result dictionary has the following keys:

        * ``name``: The component name
        * ``factory``: The name of the component factory
        * ``bundle_id``: The ID of the bundle providing the component factory
        * ``state``: The current component state
        * ``services``: A ``{Service ID → Service reference}`` dictionary, with
          all services provided by the component
        * ``dependencies``: A dictionary associating field names with the
          following dictionary:

          * ``handler``: The name of the type of the dependency handler
          * ``filter`` (optional): The requirement LDAP filter
          * ``optional``: A flag indicating whether the requirement is optional
            or not
          * ``aggregate``: A flag indicating whether the requirement is a set
            of services or not
          * ``binding``: A list of the ServiceReference the component is bound
            to

        * ``properties``: A dictionary key → value, with all properties of the
          component. The value is converted to its string representation, to
          avoid unexpected behaviours.

        :param name: The name of a component instance
        :return: A dictionary of details
        :raise ValueError: Invalid component name
        """
        if not is_string(name):
            raise ValueError("Component name must be a string")

        with self.__instances_lock:
            if name not in self.__instances:
                raise ValueError("Unknown component: {0}".format(name))

            stored_instance = self.__instances[name]
            assert isinstance(stored_instance, StoredInstance)
            with stored_instance._lock:
                result = {}  # type: Dict[str, Any]
                result["name"] = stored_instance.name

                # Factory name
                result["factory"] = stored_instance.factory_name

                # Factory bundle
                result[
                    "bundle_id"
                ] = stored_instance.bundle_context.get_bundle().get_bundle_id()

                # Component state
                result["state"] = stored_instance.state

                # Error details
                result["error_trace"] = stored_instance.error_trace

                # Provided service
                result["services"] = {}
                for handler in stored_instance.get_handlers(
                    handlers_const.KIND_SERVICE_PROVIDER
                ):
                    svc_ref = handler.get_service_reference()
                    if svc_ref is not None:
                        svc_id = svc_ref.get_property(SERVICE_ID)
                        result["services"][svc_id] = svc_ref

                # Dependencies
                result["dependencies"] = {}
                for dependency in stored_instance.get_handlers(
                    handlers_const.KIND_DEPENDENCY
                ):
                    # Dependency
                    info = result["dependencies"][dependency.get_field()] = {}
                    info["handler"] = type(dependency).__name__

                    # Requirement
                    req = dependency.requirement
                    info["specification"] = req.specification
                    info["filter"] = str(req.filter) if req.filter else None
                    info["optional"] = req.optional
                    info["aggregate"] = req.aggregate

                    # Bindings
                    info["bindings"] = dependency.get_bindings()

                # Properties
                properties = stored_instance.context.properties.items()
                result["properties"] = {
                    str(key): str(value) for key, value in properties
                }

                # All done
                return result

    def get_factories(self):
        # type: () -> List[str]
        """
        Retrieves the names of the registered factories

        :return: A list of factories. Can be empty.
        """
        with self.__factories_lock:
            return sorted(self.__factories.keys())

    def get_factory_bundle(self, name):
        # type: (str) -> Bundle
        """
        Retrieves the Pelix Bundle object that registered the given factory

        :param name: The name of a factory
        :return: The Bundle that registered the given factory
        :raise ValueError: Invalid factory
        """
        with self.__factories_lock:
            try:
                factory = self.__factories[name]
            except KeyError:
                raise ValueError("Unknown factory '{0}'".format(name))
            else:
                # Bundle Context is stored in the Factory Context
                factory_context = getattr(
                    factory, constants.IPOPO_FACTORY_CONTEXT
                )
                return factory_context.bundle_context.get_bundle()

    def get_factory_details(self, name):
        # type: (str) -> Dict[str, Any]
        """
        Retrieves a dictionary with details about the given factory

        * ``name``: The factory name
        * ``bundle``: The Bundle object of the bundle providing the factory
        * ``properties``: Copy of the components properties defined by the
          factory
        * ``requirements``: List of the requirements defined by the factory

          * ``id``: Requirement ID (field where it is injected)
          * ``specification``: Specification of the required service
          * ``aggregate``: If True, multiple services will be injected
          * ``optional``: If True, the requirement is optional

        * ``services``: List of the specifications of the services provided by
          components of this factory
        * ``handlers``: Dictionary of the non-built-in handlers required by
          this factory.
          The dictionary keys are handler IDs, and it contains a tuple with:

          * A copy of the configuration of the handler (0)
          * A flag indicating if the handler is present or not

        :param name: The name of a factory
        :return: A dictionary describing the factory
        :raise ValueError: Invalid factory
        """
        with self.__factories_lock:
            try:
                factory = self.__factories[name]
            except KeyError:
                raise ValueError("Unknown factory '{0}'".format(name))

            context = getattr(factory, constants.IPOPO_FACTORY_CONTEXT)
            assert isinstance(context, FactoryContext)

            result = {}  # type: Dict[Any, Any]
            # Factory name & bundle
            result["name"] = context.name
            result["bundle"] = context.bundle_context.get_bundle()

            # Configurable properties
            # Name -> Default value
            result["properties"] = {
                prop_name: context.properties.get(prop_name)
                for prop_name in context.properties_fields.values()
            }

            # Requirements (list of dictionaries)
            reqs = result["requirements"] = []
            handler_requires = context.get_handler(constants.HANDLER_REQUIRES)
            if handler_requires is not None:
                for field, requirement in handler_requires.items():
                    reqs.append(
                        {
                            "id": field,
                            "specification": requirement.specification,
                            "aggregate": requirement.aggregate,
                            "optional": requirement.optional,
                            "filter": requirement.original_filter,
                        }
                    )

            # Provided services (list of list of specifications)
            handler_provides = context.get_handler(constants.HANDLER_PROVIDES)
            if handler_provides is not None:
                result["services"] = [
                    specs_controller[0] for specs_controller in handler_provides
                ]
            else:
                result["services"] = []

            # Other handlers
            handlers = set(context.get_handlers_ids())
            handlers.difference_update(
                (
                    constants.HANDLER_PROPERTY,
                    constants.HANDLER_PROVIDES,
                    constants.HANDLER_REQUIRES,
                )
            )
            result["handlers"] = {
                handler: copy.deepcopy(context.get_handler(handler))
                for handler in handlers
            }

            return result


# ------------------------------------------------------------------------------


@BundleActivator
class _IPopoActivator(object):
    """
    The iPOPO bundle activator for Pelix
    """

    def __init__(self):
        """
        Sets up the activator
        """
        self._registration = None
        self._service = None
        self._bundles = []

    def start(self, context):
        # type: (BundleContext) -> None
        # pylint: disable=W0212
        """
        The bundle has started

        :param context: The bundle context
        """
        # Automatically install handlers bundles
        for handler in BUILTIN_HANDLERS:
            try:
                bundle = context.install_bundle(handler)
                bundle.start()
                self._bundles.append(bundle)
            except BundleException as ex:
                _logger.error("Error installing handler %s: %s", handler, ex)

        # Register the iPOPO service
        self._service = _IPopoService(context)
        self._registration = context.register_service(
            constants.SERVICE_IPOPO, self._service, {}
        )

        # Register as a bundle listener
        context.add_bundle_listener(self._service)

        # Register the service as a framework stop listener
        context.add_framework_stop_listener(self._service)

        # Service enters in "run" mode
        self._service.running = True

        # Get all factories
        for bundle in context.get_bundles():
            if bundle.get_state() == Bundle.ACTIVE:
                # Bundle is active, register its factories
                self._service._register_bundle_factories(bundle)

    def stop(self, context):
        # type: (BundleContext) -> None
        # pylint: disable=W0212
        """
        The bundle has stopped

        :param context: The bundle context
        """
        # The service is not in the "run" mode anymore
        self._service._stop()

        # Unregister the listener
        context.remove_bundle_listener(self._service)

        # Unregister the framework stop listener
        context.remove_framework_stop_listener(self._service)

        # Unregister the iPOPO service
        self._registration.unregister()

        # Clean up the service
        self._service._unregister_all_factories()

        # Remove handler bundles
        for bundle in self._bundles:
            bundle.uninstall()
        del self._bundles[:]

        # Clean up references
        self._registration = None
        self._service = None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: Abstract RPC implementation

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

**TODO:**
* "system" methods (list, help, ...)
"""

# Standard library
import logging
import threading
import uuid

# iPOPO decorators
from pelix.ipopo.decorators import Validate, Invalidate, Property, Provides

# Pelix constants
import pelix.constants as constants
import pelix.remote.beans
from pelix.remote import RemoteServiceError

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


@Provides(pelix.remote.SERVICE_EXPORT_PROVIDER)
@Property("_kinds", pelix.remote.PROP_REMOTE_CONFIGS_SUPPORTED)
class AbstractRpcServiceExporter(object):
    """
    Abstract Remote Services exporter
    """

    def __init__(self):
        """
        Sets up the exporter
        """
        # Bundle context
        self._context = None

        # Framework UID
        self._framework_uid = None

        # Handled configurations
        self._kinds = []

        # Exported services: Name -> ExportEndpoint
        self.__endpoints = {}

        # Thread safety
        self.__lock = threading.Lock()

    def dispatch(self, method, params):
        """
        Called by the servlet: calls the method of an exported service
        """
        # Get the best matching name
        matching = None
        len_found = 0
        for name in self.__endpoints:
            if len(name) > len_found and method.startswith(name + "."):
                # Better matching end point name (longer that previous one)
                matching = name
                len_found = len(matching)

        if matching is None:
            # No end point name match
            raise RemoteServiceError(
                "No end point found for: {0}".format(method)
            )

        # Extract the method name. (+1 for the trailing dot)
        method_name = method[len_found + 1 :]

        # Get the service
        try:
            service = self.__endpoints[matching].instance
        except KeyError:
            raise RemoteServiceError("Unknown endpoint: {0}".format(matching))

        # Get the method
        method_ref = getattr(service, method_name, None)
        if method_ref is None:
            raise RemoteServiceError("Unknown method {0}".format(method))

        # Call it (let the errors be propagated)
        if isinstance(params, (list, tuple)):
            return method_ref(*params)

        return method_ref(**params)

    def handles(self, configurations):
        """
        Checks if this provider handles the given configuration types

        :param configurations: Configuration types
        """
        if configurations is None or configurations == "*":
            # 'Matches all'
            return True

        return bool(set(configurations).intersection(self._kinds))

    def export_service(self, svc_ref, name, fw_uid):
        """
        Prepares an export endpoint

        :param svc_ref: Service reference
        :param name: Endpoint name
        :param fw_uid: Framework UID
        :return: An ExportEndpoint bean
        :raise NameError: Already known name
        :raise BundleException: Error getting the service
        """
        with self.__lock:
            # Prepare extra properties
            extra_props = self.make_endpoint_properties(svc_ref, name, fw_uid)

            try:
                # Check if the name has been changed by the exporter
                name = extra_props[pelix.remote.PROP_ENDPOINT_NAME]
            except KeyError:
                # Name not updated
                pass

            if name in self.__endpoints:
                # Already known end point
                raise NameError(
                    "Already known end point {0} for kinds {1}".format(
                        name, ",".join(self._kinds)
                    )
                )

            # Get the service (let it raise a BundleException if any
            service = self._context.get_service(svc_ref)

            # Prepare the export endpoint
            try:
                endpoint = pelix.remote.beans.ExportEndpoint(
                    str(uuid.uuid4()),
                    fw_uid,
                    self._kinds,
                    name,
                    svc_ref,
                    service,
                    extra_props,
                )
            except ValueError:
                # No specification to export (specifications filtered, ...)
                return None

            # Store information
            self.__endpoints[name] = endpoint

            # Return the endpoint bean
            return endpoint

    def update_export(self, endpoint, new_name, old_properties):
        # pylint: disable=W0613
        """
        Updates an export endpoint

        :param endpoint: An ExportEndpoint bean
        :param new_name: Future endpoint name
        :param old_properties: Previous properties
        :raise NameError: Rename refused
        """
        with self.__lock:
            try:
                if self.__endpoints[new_name] is not endpoint:
                    # Reject the new name, as an endpoint uses it
                    raise NameError(
                        "New name of {0} already used: {1}".format(
                            endpoint.name, new_name
                        )
                    )
                else:
                    # Name hasn't changed
                    pass
            except KeyError:
                # Update the name of the endpoint
                old_name = endpoint.name
                endpoint.rename(new_name)

                # No endpoint matches the new name: update the storage
                self.__endpoints[new_name] = self.__endpoints.pop(old_name)

    def unexport_service(self, endpoint):
        """
        Deletes an export endpoint

        :param endpoint: An ExportEndpoint bean
        """
        with self.__lock:
            # Clean up storage
            del self.__endpoints[endpoint.name]

            # Release the service
            svc_ref = endpoint.reference
            self._context.unget_service(svc_ref)

    def make_endpoint_properties(self, svc_ref, name, fw_uid):
        """
        Prepare properties for the ExportEndpoint to be created

        :param svc_ref: Service reference
        :param name: Endpoint name
        :param fw_uid: Framework UID
        :return: A dictionary of extra endpoint properties
        """
        raise NotImplementedError(
            "make_endpoint_properties() not "
            "implemented by class {0}".format(type(self).__name__)
        )

    @Validate
    def validate(self, context):
        """
        Component validated
        """
        # Store the context
        self._context = context

        # Store the framework UID
        self._framework_uid = context.get_property(constants.FRAMEWORK_UID)

    @Invalidate
    def invalidate(self, _):
        """
        Component invalidated
        """
        # Clean up the storage
        self.__endpoints.clear()

        # Clean up members
        self._context = None
        self._framework_uid = None


# ------------------------------------------------------------------------------


@Provides(pelix.remote.SERVICE_IMPORT_ENDPOINT_LISTENER)
@Property("_kinds", pelix.remote.PROP_REMOTE_CONFIGS_SUPPORTED)
class AbstractRpcServiceImporter(object):
    """
    Abstract Remote Services importer
    """

    def __init__(self):
        """
        Sets up the exporter
        """
        # Bundle context
        self._context = None

        # Framework UID
        self._framework_uid = None

        # Component properties
        self._kinds = None

        # Registered services (endpoint UID -> ServiceReference)
        self.__registrations = {}
        self.__lock = threading.Lock()

    def endpoint_added(self, endpoint):
        """
        An end point has been imported
        """
        configs = set(endpoint.configurations)
        if "*" not in configs and not configs.intersection(self._kinds):
            # Not for us
            return

        with self.__lock:
            if endpoint.uid in self.__registrations:
                # Already known endpoint
                return

            # Prepare a proxy
            svc = self.make_service_proxy(endpoint)
            if svc is None:
                return

            # Register it as a service
            svc_reg = self._context.register_service(
                endpoint.specifications, svc, endpoint.properties
            )

            # Store references
            self.__registrations[endpoint.uid] = svc_reg

    def endpoint_updated(self, endpoint, old_properties):
        # pylint: disable=W0613
        """
        An end point has been updated
        """
        with self.__lock:
            try:
                # Update service registration properties
                self.__registrations[endpoint.uid].set_properties(
                    endpoint.properties
                )

            except KeyError:
                # Unknown end point
                return

    def endpoint_removed(self, endpoint):
        """
        An end point has been removed
        """
        with self.__lock:
            try:
                # Pop reference and unregister the service
                self.__registrations.pop(endpoint.uid).unregister()

            except KeyError:
                # Unknown end point
                return

            else:
                # Clear the proxy
                self.clear_service_proxy(endpoint)

    def make_service_proxy(self, endpoint):
        """
        Creates the proxy for the given ImportEndpoint

        :param endpoint: An ImportEndpoint bean
        :return: A service proxy
        """
        raise NotImplementedError(
            "make_service_proxy() not implemented by "
            "class {0}".format(type(self).__name__)
        )

    def clear_service_proxy(self, endpoint):
        """
        Destroys the proxy made for the given ImportEndpoint

        :param endpoint: An ImportEndpoint bean
        """
        raise NotImplementedError(
            "clear_service_proxy() not implemented by "
            "class {0}".format(type(self).__name__)
        )

    @Validate
    def validate(self, context):
        """
        Component validated
        """
        # Store the bundle context and the framework UID
        self._context = context
        self._framework_uid = context.get_property(constants.FRAMEWORK_UID)

    @Invalidate
    def invalidate(self, _):
        """
        Component invalidated
        """
        # Unregister all of our services
        for svc_reg in self.__registrations.values():
            svc_reg.unregister()

        # Clean up members
        self.__registrations.clear()
        self._context = None
        self._framework_uid = None



"""
Pelix remote services: EDEF file handler

Endpoint Description Extender Format (EDEF) is specified in OSGi Compendium
specifications, section 122.8.

:author: Thomas Calmant and Scott Lewis
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import sys
import xml.etree.ElementTree as ElementTree

try:
    # Python 2
    from StringIO import StringIO
except ImportError:
    # Python 3
    from io import StringIO

# Typing
try:
    # pylint: disable=W0611
    from typing import Iterable, List, Tuple, Any, Union
except ImportError:
    pass

# Pelix
from pelix.rsa.endpointdescription import EndpointDescription
from pelix import rsa
import pelix.constants

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------

# EDEF XML name space
EDEF_NAMESPACE = "http://www.osgi.org/xmlns/rsa/v1.0.0"

# EDEF tags
TAG_ENDPOINT_DESCRIPTIONS = "{{{0}}}endpoint-descriptions".format(
    EDEF_NAMESPACE
)
TAG_ENDPOINT_DESCRIPTION = "{{{0}}}endpoint-description".format(EDEF_NAMESPACE)
TAG_PROPERTY = "{{{0}}}property".format(EDEF_NAMESPACE)
TAG_ARRAY = "{{{0}}}array".format(EDEF_NAMESPACE)
TAG_LIST = "{{{0}}}list".format(EDEF_NAMESPACE)
TAG_SET = "{{{0}}}set".format(EDEF_NAMESPACE)
TAG_XML = "{{{0}}}xml".format(EDEF_NAMESPACE)
TAG_VALUE = "{{{0}}}value".format(EDEF_NAMESPACE)

# Property attributes
ATTR_NAME = "name"
ATTR_VALUE_TYPE = "value-type"
ATTR_VALUE = "value"

# Value types
TYPE_BOOLEAN = "Boolean"
TYPE_DOUBLE = "Double"
TYPE_LONG = "Long"
TYPE_STRING = "String"

TYPES_BOOLEAN = ("boolean", "Boolean")
TYPES_CHAR = ("char", "Character")
TYPES_FLOAT = ("float", "Float", "double", "Double")
TYPES_INT = (
    "int",
    "Integer",
    "long",
    "Long",
    "short",
    "Short",
    "bytes",
    "Bytes",
)

# Type of properties
TYPED_BOOL = tuple()
TYPED_LONG = (
    rsa.ENDPOINT_SERVICE_ID,
    rsa.ECF_ENDPOINT_TIMESTAMP,
    rsa.ECF_RSVC_ID,
)
TYPED_STRING = (
    pelix.constants.OBJECTCLASS,
    rsa.ENDPOINT_FRAMEWORK_UUID,
    rsa.ENDPOINT_ID,
    rsa.ENDPOINT_PACKAGE_VERSION_,
    rsa.SERVICE_IMPORTED_CONFIGS,
    rsa.SERVICE_INTENTS,
    rsa.SERVICE_IMPORTED,
)

# Special case: XML value given
XML_VALUE = object()

# ------------------------------------------------------------------------------


class EDEFReader(object):
    """
    Reads an EDEF XML data. Inspired from EndpointDescriptionParser from ECF
    """

    @staticmethod
    def _convert_value(vtype, value):
        # type: (str, str) -> Any
        """
        Converts the given value string according to the given type

        :param vtype: Type of the value
        :param value: String form of the value
        :return: The converted value
        :raise ValueError: Conversion failed
        """
        # Normalize value
        value = value.strip()

        if vtype == TYPE_STRING:
            # Nothing to do
            return value
        elif vtype in TYPES_INT:
            return int(value)
        elif vtype in TYPES_FLOAT:
            return float(value)
        elif vtype in TYPES_BOOLEAN:
            # Compare lower-case value
            return value.lower() not in ("false", "0")
        elif vtype in TYPES_CHAR:
            return value[0]

        # No luck
        raise ValueError("Unknown value type: {0}".format(vtype))

    def _parse_description(self, node):
        # type: (ElementTree.Element) -> EndpointDescription
        """
        Parse an endpoint description node

        :param node: The endpoint description node
        :return: The parsed EndpointDescription bean
        :raise KeyError: Attribute missing
        :raise ValueError: Invalid description
        """
        endpoint = {}
        for prop_node in node.findall(TAG_PROPERTY):
            name, value = self._parse_property(prop_node)
            endpoint[name] = value

        return EndpointDescription(None, endpoint)

    def _parse_property(self, node):
        # type: (ElementTree.Element) -> Tuple[str, Any]
        """
        Parses a property node

        :param node: The property node
        :return: A (name, value) tuple
        :raise KeyError: Attribute missing
        """
        # Get information
        name = node.attrib[ATTR_NAME]
        vtype = node.attrib.get(ATTR_VALUE_TYPE, TYPE_STRING)

        # Look for a value as a single child node
        try:
            value_node = next(iter(node))
            value = self._parse_value_node(vtype, value_node)
        except StopIteration:
            # Value is an attribute
            value = self._convert_value(vtype, node.attrib[ATTR_VALUE])

        return name, value

    def _parse_value_node(self, vtype, node):
        # type: (str, ElementTree.Element) -> Any
        """
        Parses a value node

        :param vtype: The value type
        :param node: The value node
        :return: The parsed value
        """
        kind = node.tag
        if kind == TAG_XML:
            # Raw XML value
            return next(iter(node))

        elif kind == TAG_LIST or kind == TAG_ARRAY:
            # List
            return [
                self._convert_value(vtype, value_node.text)
                for value_node in node.findall(TAG_VALUE)
            ]

        elif kind == TAG_SET:
            # Set
            return set(
                self._convert_value(vtype, value_node.text)
                for value_node in node.findall(TAG_VALUE)
            )

        else:
            # Unknown
            raise ValueError("Unknown value tag: {0}".format(kind))

    def parse(self, xml_str):
        # type: (str) -> List[EndpointDescription]
        """
        Parses an EDEF XML string

        :param xml_str: An XML string
        :return: The list of parsed EndpointDescription
        """
        # Parse the document
        root = ElementTree.fromstring(xml_str)
        if root.tag != TAG_ENDPOINT_DESCRIPTIONS:
            raise ValueError("Not an EDEF XML: {0}".format(root.tag))

        # Parse content
        return [
            self._parse_description(node)
            for node in root.findall(TAG_ENDPOINT_DESCRIPTION)
        ]


# ------------------------------------------------------------------------------


class EDEFWriter(object):
    """
    EDEF XML file writer
    """

    def __init__(self, encoding="unicode", xml_declaration=True):
        # type: (str, bool) -> None
        """
        :param encoding: XML encoding
        :param xml_declaration: Add XML declaration
        """
        if sys.version_info[0] < 3 and encoding == "unicode":
            # Small trick for Python 2.7
            encoding = "UTF-8"

        self._encoding = encoding
        self._xml_declaration = xml_declaration

    def _indent(self, element, level=0, prefix="\t"):
        # type: (ElementTree.Element, int, str) -> None
        """
        In-place Element text auto-indent, for pretty printing.

        Code from: http://effbot.org/zone/element-lib.htm#prettyprint

        :param element: An Element object
        :param level: Level of indentation
        :param prefix: String to use for each indentation
        """
        element_prefix = "\n{0}".format(level * prefix)

        if element is not None:
            if not element.text or not element.text.strip():
                element.text = element_prefix + prefix

            if not element.tail or not element.tail.strip():
                element.tail = element_prefix

            # Yep, let the "element" variable be overwritten
            # pylint: disable=R1704
            for element in element:
                self._indent(element, level + 1, prefix)

            # Tail of the last child
            if not element.tail or not element.tail.strip():
                element.tail = element_prefix

        else:
            if level and (not element.tail or not element.tail.strip()):
                element.tail = element_prefix

    @staticmethod
    def _add_container(props_node, tag, container):
        # type: (ElementTree.Element, str, Iterable) -> None
        """
        Walks through the given container and fills the node

        :param props_node: A property node
        :param tag: Name of the container tag
        :param container: The container
        """
        values_node = ElementTree.SubElement(props_node, tag)
        for value in container:
            value_node = ElementTree.SubElement(values_node, TAG_VALUE)
            value_node.text = str(value)

    @staticmethod
    def _get_type(name, value):
        # type: (str, Any) -> Union[str, object]
        """
        Returns the type associated to the given name or value

        :param name: Property name
        :param value: Property value
        :return: A value type name
        """
        # Types forced for known keys
        if name in TYPED_BOOL:
            return TYPE_BOOLEAN

        elif name in TYPED_LONG:
            return TYPE_LONG

        elif name in TYPED_STRING:
            return TYPE_STRING

        # We need to analyze the content of value
        if isinstance(value, (tuple, list, set)):
            # Get the type from container content
            try:
                # Extract value
                value = next(iter(value))

            except StopIteration:
                # Empty list, can't check
                return TYPE_STRING

        # Single value
        if isinstance(value, int):
            # Integer
            return TYPE_LONG

        elif isinstance(value, float):
            # Float
            return TYPE_DOUBLE

        elif isinstance(value, type(ElementTree.Element(None))):
            # XML
            return XML_VALUE

        # Default: String
        return TYPE_STRING

    def _make_endpoint(self, root_node, endpoint):
        # type: (ElementTree.Element, EndpointDescription) -> None
        """
        Converts the given endpoint bean to an XML Element

        :param root_node: The XML root Element
        :param endpoint: An EndpointDescription bean
        """
        endpoint_node = ElementTree.SubElement(
            root_node, TAG_ENDPOINT_DESCRIPTION
        )

        for name, value in endpoint.get_properties().items():
            # Compute value type
            vtype = self._get_type(name, value)

            # Prepare the property node
            prop_node = ElementTree.SubElement(
                endpoint_node, TAG_PROPERTY, {ATTR_NAME: name}
            )

            if vtype == XML_VALUE:
                # Special case, we have to store the value as a child
                # without a value-type attribute
                prop_node.append(value)
                continue

            # Set the value type
            prop_node.set(ATTR_VALUE_TYPE, vtype)

            # Compute value node or attribute
            if isinstance(value, tuple):
                # Array
                self._add_container(prop_node, TAG_ARRAY, value)

            elif isinstance(value, list):
                # List
                self._add_container(prop_node, TAG_ARRAY, value)

            elif isinstance(value, set):
                # Set
                self._add_container(prop_node, TAG_SET, value)

            elif isinstance(value, type(root_node)):
                # XML (direct addition)
                prop_node.append(value)

            else:
                # Simple value -> Attribute
                prop_node.set(ATTR_VALUE, str(value))

    def _make_xml(self, endpoints):
        # type: (List[EndpointDescription]) -> ElementTree.Element
        """
        Converts the given endpoint description beans into an XML Element

        :param endpoints: A list of EndpointDescription beans
        :return: A string containing an XML document
        """
        root = ElementTree.Element(TAG_ENDPOINT_DESCRIPTIONS)
        for endpoint in endpoints:
            self._make_endpoint(root, endpoint)

        # Prepare pretty-printing
        self._indent(root)
        return root

    def to_string(self, endpoints):
        # type: (List[EndpointDescription]) -> str
        """
        Converts the given endpoint description beans into a string

        :param endpoints: A list of EndpointDescription beans
        :return: A string containing an XML document
        """
        # Make the ElementTree
        root = self._make_xml(endpoints)
        tree = ElementTree.ElementTree(root)

        # Force the default name space
        ElementTree.register_namespace("", EDEF_NAMESPACE)

        # Make the XML
        # Prepare a StringIO output
        output = StringIO()

        # Try to write with a correct encoding
        tree.write(
            output,
            encoding=self._encoding,
            xml_declaration=self._xml_declaration,
        )

        return output.getvalue().strip()

    def write(self, endpoints, filename):
        # type: (List[EndpointDescription], str) -> None
        """
        Writes the given endpoint descriptions to the given file

        :param endpoints: A list of EndpointDescription beans
        :param filename: Name of the file where to write the XML
        :raise IOError: Error writing the file
        """
        with open(filename, "w") as filep:
            filep.write(self.to_string(endpoints))



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Provides utility classes to develop REST-like API and to simplify the routing
of HTTP requests.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import inspect
import re
import uuid

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Callable, Dict, Pattern, Tuple
    from pelix.http import (
        AbstractHTTPServletRequest,
        AbstractHTTPServletResponse,
    )
except ImportError:
    pass

# Pelix utility methods
from pelix.utilities import get_method_arguments

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

HTTP_ROUTE_ATTRIBUTE = "__pelix_http_route__"
"""
Name of the attribute injected in methods to indicate their configuration
"""

# Same types as in Flask:
# string 	accepts any text without a slash (the default)
# int 	    accepts integers
# float 	like int but for floating point values
# path 	    like the default but also accepts slashes
# uuid 	    accepts UUID strings

# TODO: handle missing types
# any 	    matches one of the items provided

# Type name -> regex pattern
TYPE_PATTERNS = {
    "string": r"(?:[^/]+)",
    "int": r"(?:[+\-]?\d+)",
    "float": r"(?:[+\-]?\d+\.?\d*)",
    "path": r"(?:[\w\s/]+)",
    "uuid": r"(?:[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}"
    r"-[0-9a-fA-F]{12})",
}
TYPE_PATTERNS[None] = TYPE_PATTERNS["string"]


# Constant patterns
_MARKER_PATTERN = re.compile(r"<[^<>]*>")
_TYPED_MARKER_PATTERN = re.compile(r"<(\w+):?(\w+)?>")


def path_filter(path):
    # type: (str) -> str
    """
    Removes the trailing '/' of a path, if any

    :param path: A parsed path
    :return: The parsed path without its trailing /
    """
    return path[:-1] if path and path[-1] == "/" else path


# Type name -> conversion method (for types other than str)
TYPE_CONVERTERS = {
    "int": int,
    "float": float,
    "path": path_filter,
    "uuid": uuid.UUID,
}  # type: Dict[str, Callable[[str], Any]]

# ------------------------------------------------------------------------------


class Http(object):
    """
    Decorator indicating which route a method handles
    """

    def __init__(self, route, methods=None):
        """
        :param route: Path handled by the method (beginning with a '/')
        :param methods: List of HTTP methods allowed (GET, POST, ...)
        """
        if not route:
            # Normalize route
            route = "/"
        else:
            # Remove surrounding spaces
            route = route.strip()

        if methods and not isinstance(methods, (list, tuple, set, frozenset)):
            # Normalize methods
            raise TypeError("methods should be a list")

        self._route = route
        self._methods = methods or ["GET"]

    def __call__(self, decorated_method):
        """
        Injects the HTTP_ROUTE_ATTRIBUTE to the decorated method to store the
        description of the route

        :param decorated_method: The decorated method
        """
        if not inspect.isroutine(decorated_method):
            raise TypeError(
                "@Http can decorate only methods, not {0}".format(
                    type(decorated_method).__name__
                )
            )

        try:
            config = getattr(decorated_method, HTTP_ROUTE_ATTRIBUTE)
        except AttributeError:
            config = {}
            setattr(decorated_method, HTTP_ROUTE_ATTRIBUTE, config)

        # Use sets to avoid duplications
        config.setdefault("routes", set()).add(self._route)
        config.setdefault("methods", set()).update(self._methods)
        return decorated_method


class HttpGet(Http):
    """
    Decorates a method handling GET requests
    """

    def __init__(self, route):
        """
        :param route: Path handled by the method (beginning with a '/')
        """
        super(HttpGet, self).__init__(route, methods=["GET"])


class HttpHead(Http):
    """
    Decorates a method handling HEAD requests
    """

    def __init__(self, route):
        """
        :param route: Path handled by the method (beginning with a '/')
        """
        super(HttpHead, self).__init__(route, methods=["HEAD"])


class HttpPost(Http):
    """
    Decorates a method handling POST requests
    """

    def __init__(self, route):
        """
        :param route: Path handled by the method (beginning with a '/')
        """
        super(HttpPost, self).__init__(route, methods=["POST"])


class HttpPut(Http):
    """
    Decorates a method handling PUT requests
    """

    def __init__(self, route):
        """
        :param route: Path handled by the method (beginning with a '/')
        """
        super(HttpPut, self).__init__(route, methods=["PUT"])


class HttpDelete(Http):
    """
    Decorates a method handling DELETE requests
    """

    def __init__(self, route):
        """
        :param route: Path handled by the method (beginning with a '/')
        """
        super(HttpDelete, self).__init__(route, methods=["DELETE"])


# ------------------------------------------------------------------------------


class RestDispatcher(object):
    """
    Parent class for servlets: dispatches requests according to the @Http
    decorator
    """

    def __init__(self):
        """
        Looks for the methods where to dispatch requests
        """
        # HTTP verb -> route pattern -> function
        self.__routes = {}

        # function -> arg name -> arg converter
        self.__methods_args = {}

        # Find all REST methods
        self._setup_rest_dispatcher()

    def do_GET(self, request, response):
        # pylint: disable=C0103
        """
        Handles a GET request
        """
        self._rest_dispatch(request, response)

    def do_HEAD(self, request, response):
        # pylint: disable=C0103
        """
        Handles a HEAD request
        """
        self._rest_dispatch(request, response)

    def do_POST(self, request, response):
        # pylint: disable=C0103
        """
        Handles a POST request
        """
        self._rest_dispatch(request, response)

    def do_PUT(self, request, response):
        # pylint: disable=C0103
        """
        Handles a PUT request
        """
        self._rest_dispatch(request, response)

    def do_DELETE(self, request, response):
        # pylint: disable=C0103
        """
        Handles a DELETE request
        """
        self._rest_dispatch(request, response)

    def _rest_dispatch(self, request, response):
        # type: (AbstractHTTPServletRequest, AbstractHTTPServletResponse) -> None
        """
        Dispatches the request

        :param request: Request bean
        :param response: Response bean
        """
        # Extract request information
        http_verb = request.get_command()
        sub_path = request.get_sub_path()

        # Find the best matching method, according to the number of
        # readable arguments
        max_valid_args = -1
        best_method = None
        best_args = None
        best_match = None

        for route, method in self.__routes.get(http_verb, {}).items():
            # Parse the request path
            match = route.match(sub_path)
            if not match:
                continue

            # Count the number of valid arguments
            method_args = self.__methods_args[method]
            nb_valid_args = 0
            for name in method_args:
                try:
                    match.group(name)
                    nb_valid_args += 1
                except IndexError:
                    # Argument not found
                    pass

            if nb_valid_args > max_valid_args:
                # Found a better match
                max_valid_args = nb_valid_args
                best_method = method
                best_args = method_args
                best_match = match

        if best_method is None:
            # No match: return a 404 plain text error
            response.send_content(
                404,
                "No method to handle path {0}".format(sub_path),
                "text/plain",
            )
        else:
            # Found a method
            # ... convert arguments
            kwargs = {}
            if best_args:
                for name, converter in best_args.items():
                    try:
                        str_value = best_match.group(name)
                    except IndexError:
                        # Argument is missing: do nothing
                        pass
                    else:
                        if str_value:
                            # Keep the default value when an argument is
                            # missing, i.e. don't give it in kwargs
                            if converter is not None:
                                # Convert the argument
                                kwargs[name] = converter(str_value)
                            else:
                                # Use the string value as is
                                kwargs[name] = str_value

            # Prepare positional arguments
            extra_pos_args = []
            if kwargs:
                # Ignore the first two parameters (request and response)
                method_args = get_method_arguments(best_method).args[:2]
                for pos_arg in method_args:
                    try:
                        extra_pos_args.append(kwargs.pop(pos_arg))
                    except KeyError:
                        pass

            # ... call the method (exceptions will be handled by the server)
            best_method(request, response, *extra_pos_args, **kwargs)

    def _setup_rest_dispatcher(self):
        """
        Finds all methods to call when handling a route
        """
        for _, method in inspect.getmembers(self, inspect.isroutine):
            try:
                config = getattr(method, HTTP_ROUTE_ATTRIBUTE)
            except AttributeError:
                # Not a REST method
                continue

            for route in config["routes"]:
                pattern, arguments = self.__convert_route(route)
                self.__methods_args.setdefault(method, {}).update(arguments)
                for http_verb in config["methods"]:
                    self.__routes.setdefault(http_verb, {})[pattern] = method

    @staticmethod
    def __convert_route(route):
        # type: (str) -> Tuple[Pattern[str], Dict[str, Callable[[str], Any]]]
        """
        Converts a route pattern into a regex.
        The result is a tuple containing the regex pattern to match and a
        dictionary associating arguments names and their converter (if any)

        A route can be: "/hello/<name>/<age:int>"

        :param route: A route string, i.e. a path with type markers
        :return: A tuple (pattern, {argument name: converter})
        """
        arguments = {}  # type: Dict[str, Callable[[str], Any]]
        last_idx = 0
        final_pattern = []
        match_iter = _MARKER_PATTERN.finditer(route)
        for match_pattern in match_iter:
            # Copy intermediate string
            final_pattern.append(route[last_idx : match_pattern.start()])
            last_idx = match_pattern.end() + 1

            # Extract type declaration
            match_type = _TYPED_MARKER_PATTERN.match(match_pattern.group())
            if not match_type:
                raise ValueError(
                    "Invalid argument declaration: {0}".format(
                        match_pattern.group()
                    )
                )

            name, kind = match_type.groups()
            if kind:
                kind = kind.lower()

            # Choose a pattern for each type (can raise a KeyError)
            regex = TYPE_PATTERNS[kind]

            # Keep track of argument name and converter
            arguments[name] = TYPE_CONVERTERS.get(kind)

            # Generate the regex pattern for this part
            final_pattern.append("((?P<")
            final_pattern.append(match_type.group(1))
            final_pattern.append(">")
            final_pattern.append(regex)
            final_pattern.append(")/?)?")

        # Copy trailing string
        final_pattern.append(route[last_idx:])

        # Ensure we don't accept trailing values
        final_pattern.append("$")
        return re.compile("".join(final_pattern)), arguments
